"""Gathers the formulas for the activity of LiCl solutions.

Sources
-------
- Conde, M. R., Properties of aqueous solutions of lithium and calcium
  chlorides: formulations for use in air conditioning equipment design.
  International Journal of Thermal Sciences 43, 367-382 (2004).

- Pitzer, K. S. & Mayorga, G.
  Thermodynamics of electrolytes. II.
  Activity and osmotic coefficients for strong electrolytes with one or both
  ions univalent.
  J. Phys. Chem. 77, 2300-2308
  (1973)
"""

# TODO: add Gibbard 1973 ?

from .conde import WaterActivity_LiCl_Conde_Base
from .patil import WaterActivity_LiCl_Patil_Base
from .pitzer import WaterActivity_LiCl_Pitzer_Base


class WaterActivity_LiCl_Conde(WaterActivity_LiCl_Conde_Base):
    """Already defined in conde module"""
    default = True


class WaterActivity_LiCl_Patil(WaterActivity_LiCl_Patil_Base):
    """Already defined in pitzer module"""
    pass


class WaterActivity_LiCl_Pitzer(WaterActivity_LiCl_Pitzer_Base):
    """Already defined in pitzer module"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

WaterActivityFormulas_LiCl = (
    WaterActivity_LiCl_Conde,
    WaterActivity_LiCl_Patil,
    WaterActivity_LiCl_Pitzer,
)



"""Activity of solutions according to Pitzer original paper.

NOTE: Almost identical in structure to activity_coefficient.pitzer

Sources
-------
Patil, K. R., Tripathi, A. D., Pathak, G. & Katti, S. S.
Thermodynamic Properties of Aqueous Electrolyte Solutions.
1. Vapor Pressure of Aqueous Solutions of LiCI, LiBr, and LiI.
J. Chem. Eng. Data 35, 166-168
(1990)
"""

# TODO: Patil has a correction for second virial coefficient of water that I
# have not included here (B_T)

# TODO: Add coefficients for LiI

import numpy as np

from ...water.vapor_pressure import VaporPressure_IAPWS
from ...general import SolutionFormula


class WaterActivity_Patil_Base(SolutionFormula):

    source = 'Patil'

    temperature_unit = 'K'
    temperature_range = (298.15, 348.15) # Extended 5°C from their meas. range

    concentration_unit = 'm'

    with_water_reference = False

    def _coeff(self, m, coeffs):
        """Common calculations for coeffs A, B, C as a function of m"""
        X0, X1, X2, X3 = coeffs
        return X0 + X1 * m + X2 * m**2 + X3 * m**3

    def calculate(self, m, T):
        A = self._coeff(m, coeffs=self.coeffs['A'])
        B = self._coeff(m, coeffs=self.coeffs['B'])
        C = self._coeff(m, coeffs=self.coeffs['C'])
        log_p_kPa = A + B / T + C / T**2
        p = 1e3 * 10**log_p_kPa
        vapor_pressure = VaporPressure_IAPWS()
        psat = vapor_pressure.calculate(T=T)
        return p / psat


# ============================ Different solutes =============================


class WaterActivity_LiCl_Patil_Base(WaterActivity_Patil_Base):
    solute = 'LiCl'
    concentration_range = (3, 18.5)
    coeffs = {
        'A': (7.323_3550, -0.062_3661, 0.006_1613, -0.000_1042),
        'B': (-1718.1570, 8.2255, -2.2131, 0.0246),
        'C': (-97_575.680, 3839.979, -421.429, 16.731),
    }


class WaterActivity_LiBr_Patil_Base(WaterActivity_Patil_Base):
    solute = 'LiBr'
    concentration_range = (2, 16)
    coeffs = {
        'A': (8.064_8240, -0.103_6791, -0.012_3511, 0.000_9890),
        'B': (-2235.3810, 62.4534, 5.2484, -0.5253),
        'C': (-6478.216, -10_555.860, -724.251, 66.490),
    }



"""Gathers the formulas for the solubility of Na2SO4.


Sources
-------
- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

(some info of domain of validity of expressions in the following paper:)
Dorn, J. & Steiger, M. Measurement and Calculation of Solubilities in the
Ternary System NaCH 3 COO + NaCl + H 2 O from 278 K to 323 K.
J. Chem. Eng. Data 52, 1784-1790 (2007).)
"""

from .steiger import Solubility_Na2SO4_Steiger2008_Base
from .steiger import Solubility_Na2SO4_10H2O_Steiger2008_Base


class Solubility_Na2SO4_Steiger(Solubility_Na2SO4_Steiger2008_Base):
    """Already defined in steiger module. Default formula for thenardite."""
    default = True


class Solubility_Na2SO4_10H2O_Steiger(Solubility_Na2SO4_10H2O_Steiger2008_Base):
    """Already defined in steiger module. Default formula for mirabilite"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

SolubilityFormulas_Na2SO4 = (
    Solubility_Na2SO4_Steiger,
    Solubility_Na2SO4_10H2O_Steiger,
)



"""Surface tension of solutions from Conde 2004

Source
------
Conde, M. R., Properties of aqueous solutions of lithium and calcium
chlorides: formulations for use in air conditioning equipment design.
International Journal of Thermal Sciences 43, 367-382 (2004).
"""

from ....constants import Tc
from ...general import SolutionFormula
from .misc import sigma_iapws


def sigma_conde(w, T, coeffs):
    """General formula for surface tension accorging to Conde IJTS 2004.

    Inputs
    ------
    w: weight fraction of salt
    T: temperature in K
    coeffs: coefficients sigma_i from Table 6
    coeffs_table5: tuple or list of coefficients aws, bws, asw, bsw (table 5)

    Outputs
    -------
    Surface tension of solution, N/m

    Notes
    -----
    Ranges of validity for temperature and concentration are given in Table 2.

    Reference
    ---------
    Conde, M. R., Properties of aqueous solutions of lithium and calcium
    chlorides: formulations for use in air conditioning equipment design.
    International Journal of Thermal Sciences 43, 367-382 (2004).
    """

    # surface tension of pure water
    sigma_w = sigma_iapws(T)

    # surface tension of the solution
    t = T / Tc
    s1, s2, s3, s4, s5 = coeffs
    r = 1 + s1 * w + s2 * w * t + s3 * w * t ** 2 + s4 * w ** 2 + s5 * w ** 3

    return sigma_w * r


class SurfaceTension_Conde_Base(SolutionFormula):

    source ='Conde'

    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)

    concentration_unit = 'w'
    concentration_range = (0, 0.45)

    with_water_reference = True

    def calculate(self, w, T):
        """Surface tension calculated from Conde 2004.
        Input: weight fraction w, temperature T in K."""
        sigma_w = sigma_iapws(T)
        sigma = sigma_conde(w, T, self.coeffs)
        return sigma_w, sigma


class SurfaceTension_CaCl2_Conde_Base(SurfaceTension_Conde_Base):
    solute = 'CaCl2'
    coeffs = [2.33067, -10.78779, 13.56611, 1.95017, -1.77990]


class SurfaceTension_LiCl_Conde_Base(SurfaceTension_Conde_Base):
    solute = 'LiCl'
    coeffs = [2.757115, -12.011299, 14.751818, 2.443204, -3.147739]



"""Activity of solutions from Conde 2004

Source
------
- Conde, M. R., Properties of aqueous solutions of lithium and calcium
chlorides: formulations for use in air conditioning equipment design.
International Journal of Thermal Sciences 43, 367-382 (2004).
"""

import numpy as np

from ....constants import Tc
from ...general import SolutionFormula


def aw_conde(w, T, coeffs):
    """General formula for surface tension accorging to Conde IJTS 2004.

    Inputs
    ------
    w: weight fraction of salt
    T: temperature in K
    coeffs: coeffs pi_i from Table 3

    Outputs
    -------
    Water activity of solution (dimensionless)

    Notes
    -----
    The expression does not tend to 1 at 0 concentration and is not defined
    for w=0 !!

    Reference
    ---------
    Conde, M. R., Properties of aqueous solutions of lithium and calcium
    chlorides: formulations for use in air conditioning equipment design.
    International Journal of Thermal Sciences 43, 367-382 (2004).
    """
    pi0, pi1, pi2, pi3, pi4, pi5, pi6, pi7, pi8, pi9 = coeffs

    a = 2 - (1 + (w / pi0)**pi1)**pi2
    b = (1 + (w / pi3)**pi4)**pi5 - 1

    t = T / Tc

    f = a + b * t
    pi25 = 1 - (1 + (w / pi6)**pi7)**pi8 - pi9 * np.exp(-(w - 0.1)**2 / 0.005)

    return f * pi25


class WaterActivity_Conde_Base(SolutionFormula):

    source ='Conde'
    solute = 'CaCl2'

    temperature_unit = 'C'
    temperature_range = (0, 100)   # Deduced from data presented in Fig. 3

    concentration_unit = 'w'

    with_water_reference = False

    def calculate(self, w, T):
        """Water activity for LiCl as a function of concentration according to Conde."""
        T = T + 273.15
        aw = aw_conde(w, T, self.coeffs)
        return aw


class WaterActivity_CaCl2_Conde_Base(WaterActivity_Conde_Base):
        solute = 'CaCl2'
        # Approximative, actually depends on temperature. Conde not defined in w=0 ...
        concentration_range = (1e-9, 0.6)
        coeffs = [0.31, 3.698, 0.6, 0.231, 4.584, 0.49, 0.478, -5.20, -0.4, 0.018]


class WaterActivity_LiCl_Conde_Base(WaterActivity_Conde_Base):
        solute = 'LiCl'
        # Approximative, actually depends on temperature. Conde not defined in w=0 ...
        concentration_range = (1e-9, 0.55)
        coeffs = [0.28, 4.30, 0.6, 0.21, 5.1, 0.49, 0.362, -4.75, -0.4, 0.03]



"""Gathers the formulas for the refractive index of NaCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter:
- n (index of refraction, dimensionless)
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Tan (default) : "Dependence of Refractive Index on Concentration and
Temperature in Electrolyte Solution, Polar Solution, Nonpolar Solution, and
Protein Solution", Tan & Huang, J. Chem. Eng. Data  (2015).
Valid from w = 0 to w = 0.25 and for temperatures between 20 and 45°C
"""

from .tan import RefractiveIndex_NaCl_Tan_Base


class RefractiveIndex_NaCl_Tan(RefractiveIndex_NaCl_Tan_Base):
    """Already defined in tan module"""
    default = True


# ============================= WRAP-UP FORMULAS =============================

RefractiveIndexFormulas_NaCl = (
    RefractiveIndex_NaCl_Tan,
)

# ====================== DIRECT RUN (test of formulas) =======================

if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import numpy as np

    from aquasol.solutions import refractive_index

    # This shows that the quadratic term in the Tan paper for concentration is
    # not correct and should be ignored

    ws = 0.01, 0.05, 0.10, 0.15, 0.20, 0.25  # experimental weight fraction
    ww = np.linspace(0, 0.25, 100)           # weight fraction for fit

    nexps = {25: [1.3359, 1.3428, 1.3514, 1.3604, 1.3692, 1.3788], # 25°C
             45: [1.3333, 1.3398, 1.3485, 1.3572, 1.3661, 1.3757]} # 45°C

    fig, ax = plt.subplots()

    for T, ns in nexps.items():

        nn = refractive_index(w=ww, T=T)

        ax.plot(ws, ns, '.')
        ax.plot(ww, nn, '-')

    plt.show()



"""Function to calculate the viscosity of liquid water as a function of T.

Sources
-------

--- 'IAPWS':
    Huber, M. L. et al.
    "New International Formulation for the Viscosity of H2O."
    Journal of Physical and Chemical Reference Data 38, 101-125
    (2009)
"""

from ..general import WaterFormula


class ViscosityAtm_IAPWS(WaterFormula):

    source ='IAPWS'
    temperature_unit = 'K'
    temperature_range = (253.15, 383.15)
    default = True

    def calculate(self, T):
        """Viscosity of liquid water according to Huber 2009 (IAPWS)

        Input
        -----
        Temperature in K

        Output
        ------
        Viscosity in Pa.s

        Reference
        ---------
        Huber, M. L. et al.
        New International Formulation for the Viscosity of H2O.
        Journal of Physical and Chemical Reference Data 38, 101-125 (2009).

        Notes
        -----
        - Valid between 253.15 K and 383.15 K (metastable domains included)
        """
        t = T / 300

        mu_1 = 280.68 * t ** (-1.9)
        mu_2 = 511.45 * t ** (-7.7)
        mu_3 = 61.131 * t ** (-19.6)
        mu_4 = .45903 * t ** (-40)

        return (mu_1 + mu_2 + mu_3 + mu_4) * 1e-6


# ========================== WRAP-UP OF FORMULAS =============================

ViscosityAtmFormulas = (
    ViscosityAtm_IAPWS,
)



"""Density of solutions"""

from .CaCl2 import Density_CaCl2_Formulas
from .KCl import Density_KCl_Formulas
from .KI import Density_KI_Formulas
from .LiCl import Density_LiCl_Formulas
from .MgCl2 import Density_MgCl2_Formulas
from .Na2SO4 import Density_Na2SO4_Formulas
from .NaCl import Density_NaCl_Formulas

DensityFormulas = (
    Density_CaCl2_Formulas +
    Density_KCl_Formulas +
    Density_KI_Formulas +
    Density_LiCl_Formulas +
    Density_MgCl2_Formulas +
    Density_Na2SO4_Formulas +
    Density_NaCl_Formulas
)



"""Gathers the formulas for the density of NaCl solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters:
- rho0, density of pure water in kg / m^3
- rho, density of solution in kg / m^3
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------

- Simion (default) : "Mathematical modelling of density and
  viscosity of NaCl aqueous solutions" (2015). Valid from w = 0 to w = 0.26
  and for temperatures between 0 and 100°C

- Tang: "Chemical and size effects of hygroscopic aerosols on light
  scattering coefficients" (1996). Valid at 25°C and from w = 0 to w ~= 0.5

- Al Ghafri et al., Densities of Aqueous MgCl 2 (aq), CaCl 2 (aq), KI(aq),
  NaCl(aq), KCl(aq), AlCl 3 (aq), and (0.964 NaCl + 0.136 KCl)(aq) at
  Temperatures Between (283 and 472) K, Pressures up to 68.5 MPa, and
  Molalities up to 6 mol·kg -1.
  Journal of Chemical & Engineering Data 57, 1288-1304 (2012).

- Steiger:
  Talreja-Muthreja, T., Linnow, K., Enke, D. & Steiger, M.
  Deliquescence of NaCl Confined in Nanoporous Silica.
  Langmuir 38, 10963-10974 (2022).

- Krumgalz, B. S., Pogorelsky, R. & Pitzer, K. S.
  Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation
  Concentration at 298.15 °K Represented by Pitzer's Ion-Interaction Equations.
  Journal of Physical and Chemical Reference Data 25, 663-689 (1996).

- Clegg, S. L. & Wexler, A. S.
  Densities and Apparent Molar Volumes of Atmospherically Important
  Electrolyte Solutions. 1. The Solutes H2SO4, HNO3, HCl, Na2SO4, NaNO3, NaCl,
  (NH4)2SO4, NH4NO3, and NH4Cl from 0 to 50 °C, Including Extrapolations to
  Very Low Temperature and to the Pure Liquid State, and NaHSO4, NaOH, and NH3
  at 25 °C. J. Phys. Chem. A 115, 3393-3460 (2011).
"""

import numpy as np

from ...general import SolutionFormula
from ...water.density_atm import DensityAtm_IAPWS

from .clegg import density_NaCl
from .al_ghafri import Density_NaCl_AlGhafri_Base
from .krumgalz import Density_NaCl_Krumgalz_Base
from .misc import rho_tang, density_pitzer


class Density_NaCl_Simion(SolutionFormula):

    source ='Simion'
    solute = 'NaCl'

    temperature_unit = 'C'
    temperature_range = (0, 100)

    concentration_unit = 'w'
    concentration_range = (0, 0.27)

    default = True
    with_water_reference = True

    coeffs = {
        'a1': 750.2834,
        'a2': 26.7822,
        'a3': -0.26389,
        'a4': 1.90165,
        'a5': -0.11734,
        'a6': 0.00175,
        'a7': -0.003604,
        'a8': 0.0001701,
        'a9': -0.00000261,
    }

    def calculate(self, w, T):

        w = w * 100  # avoid using *= to not mutate objects in place
        T = T + 273.15  # same

        a1, a2, a3, a4, a5, a6, a7, a8, a9 = self.coeffs.values()

        rho0 = a1 + a4*T + a7*T**2  # density of pure water
        rho = rho0 + a2*w + a3*w**2 + (a5*w + a6*w**2) * T + (a8*w + a9*w**2) * T**2

        return rho0, rho

class Density_NaCl_Tang(SolutionFormula):

    source ='Tang'
    solute = 'NaCl'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'w'
    concentration_range = (0, 0.5)

    with_water_reference = True

    coeffs = np.array([7.41e-3, -3.741e-5, 2.252e-6, -2.06e-8]) * 1000

    def calculate(self, w, T):
        return rho_tang(w, self.coeffs)


class Density_NaCl_AlGhafri(Density_NaCl_AlGhafri_Base):
    """Already defined in Al Ghafri module and not default here"""
    pass


class Density_NaCl_Steiger(SolutionFormula):

    source ='Steiger'
    solute = 'NaCl'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'm'
    concentration_range = (0, 6.1)

    with_water_reference = True

    def calculate(self, m, T):
        return density_pitzer(m, solute='NaCl', source='Steiger')


class Density_NaCl_Krumgalz(Density_NaCl_Krumgalz_Base):
    """Already defined in Krumgalz module and not default here"""
    pass


class Density_NaCl_Clegg(SolutionFormula):

    source ='Clegg'
    solute = 'NaCl'

    temperature_unit = 'K'
    temperature_range = (273.15, 323.15)

    concentration_unit = 'w'
    concentration_range = (0.25, 1)

    with_water_reference = True

    def calculate(self, w, T):
        density_atm = DensityAtm_IAPWS()  # because used internally by Clegg
        rho_0 = density_atm.calculate(T)
        rho = density_NaCl(w, T)
        return rho_0, rho

# ========================== WRAP-UP OF FORMULAS =============================

Density_NaCl_Formulas = (
    Density_NaCl_Simion,
    Density_NaCl_Tang,
    Density_NaCl_AlGhafri,
    Density_NaCl_Steiger,
    Density_NaCl_Krumgalz,
    Density_NaCl_Clegg,
)


"""Electrical conductivity of solutions"""

from .KCl import ElectricalConductivity_KCl_Formulas
from .NaCl import ElectricalConductivity_NaCl_Formulas

ElectricalConductivityFormulas = (
    ElectricalConductivity_KCl_Formulas +
    ElectricalConductivity_NaCl_Formulas
)


"""Gathers the formulas for the density of Na2SO4 solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters:
- rho0, density of pure water in kg / m^3
- rho, density of solution in kg / m^3
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------

- Tang, I. N. & Munkelwitz, H. R.
  Simultaneous Determination of Refractive Index and Density of an
  Evaporating Aqueous Solution Droplet.
  Aerosol Science and Technology 15, 201-207 (1991).

- Krumgalz, B. S., Pogorelsky, R. & Pitzer, K. S.
  Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation
  Concentration at 298.15 °K Represented by Pitzer's Ion-Interaction Equations.
  Journal of Physical and Chemical Reference Data 25, 663-689 (1996).

- Clegg, S. L. & Wexler, A. S.
  Densities and Apparent Molar Volumes of Atmospherically Important
  Electrolyte Solutions. 1. The Solutes H2SO4, HNO3, HCl, Na2SO4, NaNO3, NaCl,
  (NH4)2SO4, NH4NO3, and NH4Cl from 0 to 50 °C, Including Extrapolations to
  Very Low Temperature and to the Pure Liquid State, and NaHSO4, NaOH, and NH3
  at 25 °C. J. Phys. Chem. A 115, 3393-3460 (2011).
"""

import numpy as np

from ...general import SolutionFormula
from ...water.density_atm import DensityAtm_IAPWS

from .clegg import density_Na2SO4_high_conc
from .krumgalz import Density_Na2SO4_Krumgalz_Base
from .misc import rho_tang


class Density_Na2SO4_Tang(SolutionFormula):

    source ='Tang'
    solute = 'Na2SO4'

    temperature_unit = 'C'
    temperature_range = (25, 25)

    concentration_unit = 'w'
    concentration_range = (0, 0.68)

    default = True
    with_water_reference = True

    coeffs = np.array([8.871e-3, 3.195e-5, 2.28e-7, 0]) * 1000

    def calculate(self, w, T):
        return rho_tang(w, self.coeffs)


class Density_Na2SO4_Krumgalz(Density_Na2SO4_Krumgalz_Base):
    """Already defined in Krumgalz module and not default here"""
    pass


class Density_Na2SO4_Clegg(SolutionFormula):

    source ='Clegg'
    solute = 'Na2SO4'

    temperature_unit = 'K'
    temperature_range = (273.15, 348.15)  # 0°C to 75°C

    concentration_unit = 'w'
    concentration_range = (0.22, 1)

    with_water_reference = True

    def calculate(self, w, T):
        density_atm = DensityAtm_IAPWS()
        rho_w = density_atm.calculate(T=T)
        # NOTE : there is also a formula that I coded for low concentration
        # but there seems to be a problem somewhere because it's not continuous
        # with the high concentration formula
        # (see clegg.py module)
        return rho_w, density_Na2SO4_high_conc(w, T)



# ========================== WRAP-UP OF FORMULAS =============================

Density_Na2SO4_Formulas = (
    Density_Na2SO4_Tang,
    Density_Na2SO4_Krumgalz,
    Density_Na2SO4_Clegg,
)


"""Various functions for calculation of the water activity of solutions."""


import numpy as np

from ....constants import Mw, charge_numbers, dissociation_numbers




def aw_extended_debye_huckel(m, T, solute, coeffs):
    """Mix of Hamer & Wu 1972 and Tang, Munkelwitz and Wang 1986.

    Used for NaCl and KCl
    """
    z1, z2 = charge_numbers[solute]
    nu = sum(dissociation_numbers[solute])

    A, B, C, D, E, beta = coeffs

    b = 1 + B * np.sqrt(m)

    term1 = (z1 * z2 * A / (B**3 * m)) * (b - 4.60517 * np.log10(b) - 1 / b)
    term2 = - (beta * m / 2) - (2 / 3 * C * m**2) - (3 / 4 * D * m**3) - (4 / 5 * E * m**4)

    phi =  1 - 2.302585 * (term1 + term2)  # osmotic coefficient

    return np.exp(-nu * Mw * phi * m)


"""Gathers the formulas for the activity coefficients of LiCl solutions.

Sources
-------

Pitzer, K. S. & Mayorga, G.
Thermodynamics of electrolytes. II.
Activity and osmotic coefficients for strong electrolytes with one or both
ions univalent.
J. Phys. Chem. 77, 2300-2308
(1973)
"""

from .pitzer import ActivityCoefficient_LiCl_Pitzer_Base


class ActivityCoefficient_LiCl_Pitzer(ActivityCoefficient_LiCl_Pitzer_Base):
    """Already defined in pitzer module"""
    default = True


# ========================== WRAP-UP OF FORMULAS =============================

ActivityCoefficientFormulas_LiCl = (
    ActivityCoefficient_LiCl_Pitzer,
)



"""Gathers the formulas for the surface tension of Na2SO4 solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters
- sigma0, sigma surface tensions in N/m of pure water and solution
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Dutcher: Dutcher, C. S., Wexler, A. S. & Clegg, S. L. Surface Tensions of
Inorganic Multicomponent Aqueous Electrolyte Solutions and Melts.
J. Phys. Chem. A 114, 12216-12230 (2010).
"""

from .dutcher import SufaceTension_Na2SO4_Dutcher_Base


class SurfaceTension_Na2SO4_Dutcher(SufaceTension_Na2SO4_Dutcher_Base):
    """Already defined in dutcher module"""
    default = True


# ========================== WRAP-UP OF FORMULAS =============================

SurfaceTensionFormulas_Na2SO4 = (
    SurfaceTension_Na2SO4_Dutcher,
)




"""Functions for the calculation of the density of solutions."""

import numpy as np

from ....constants import molar_mass
from ...water.density_atm import DensityAtm_IAPWS

from ..pitzer import PitzerVolumetric


# ----------------------------------------------------------------------------
# ============================ Pitzer-based formulas =========================
# ----------------------------------------------------------------------------


# ============= Molar volume Constants from Krumgalz 1996 (25°C) =============
# ----------- Note: MANY more solutes available in that reference ------------

Av_krum = 1.875

vm0_infty_krum = {
    'NaCl': 16.620,  # cm^3 / mol
    'KCl': 26.848,
    'LiCl': 16.866,
    'CaCl2': 17.612,
    'LiBr': 23.758,
    'KI': 45.151,
    'NaBr': 23.479,
    'Na2SO4': 11.766,
    'MgCl2': 14.083,
}

# Converted to kg·mol–1·MPa–1 (originally bar-1)
params_krum = {  # beta0    beta1       C
    'NaCl': (1.2335e-4, 0.43543e-4, -0.6578e-5),
    'KCl': (1.2793e-4, 0.8948e-4, -0.7131e-5),
    'LiCl': (0.3853e-4, 1.5553e-4,  -0.1541e-5),
    'CaCl2': (1.3107e-4, -2.4575e-4,  -0.1265e-5),
    'LiBr': (0.2399e-4, 0.2406e-4,  -0.1744e-5),
    'KI': (0.5398e-4, 1.4438e-4, -0.2479e-5),
    'NaBr': (0.7607e-4, 0.9525e-4, -0.3491e-5),
    'Na2SO4': (5.3250e-4, 12.932e-4, -2.914e-5),
    'MgCl2': (1.6933e-4, -5.2068e-4, -0.5698e-5),
}

# max molalities (mol/kg) for formula to be valid (cf table 2)
m_max_krum = {
    'NaCl': 6.1,
    'KCl': 4.7,
    'LiCl': 19.6,
    'CaCl2': 7.7,
    'LiBr': 17.7,
    'KI': 8.6,
    'NaBr': 8.0,
    'Na2SO4': 1.5,
    'MgCl2': 5.8,
}

# ============= Molar volume Constants from Steiger 2022 (25°C) ==============

# Slightly different from original Krumgalz (1.875, cf p. 665), taken from Archer and Wang
Av_steiger = 1.8305
vm0_infty_steiger = {'NaCl': 16.620}
params_steiger = {'NaCl': (1.17996e-4, 0.713188e-4, -0.586951e-5)}

# ================= Gather coeffs from Steiger and Krumgalz =================

Avs = {'Krumgalz': Av_krum, 'Steiger': Av_steiger}
vm0_inftys = {'Krumgalz': vm0_infty_krum, 'Steiger': vm0_infty_steiger}
params = {'Krumgalz': params_krum, 'Steiger': params_steiger}


def apparent_molar_volume_pitzer(m, solute='NaCl', source='Krumgalz'):
    """Determine apparente molar volume from Krumgaltz."""
    beta0, beta1, Cv = params[source][solute]
    v0 = vm0_inftys[source][solute]
    pitzer_v = PitzerVolumetric(T=298.15,
                                solute=solute,
                                A_v=Avs[source],
                                beta0=beta0,
                                beta1=beta1,
                                C_v=Cv,
                                v_0=v0)

    # conversion cm^3 / mol --> m^3 / mol
    return pitzer_v.apparent_molar_volume(m=m) * 1e-6


def density_pitzer(m, solute='NaCl', source='Krumgalz'):
    """Density at 25°C in kg / m^3"""
    v_phi = apparent_molar_volume_pitzer(m, solute=solute, source=source)
    Ms = molar_mass(solute)
    density_atm = DensityAtm_IAPWS()
    rho_w = density_atm.calculate(T=298.15)
    return rho_w, rho_w * (1 + m * Ms) / (1 + m * rho_w * v_phi)


# ----------------------------------------------------------------------------
# ============================== Other formulas ==============================
# ----------------------------------------------------------------------------


def relative_rho_conde(z, coeffs):
    """Relative Density of LiCl and CaCl2 according to Conde 2004.

    - z is mass fraction (w / (1 - w))
    - T is temperature in K
    - coeffs are the rho_i coefficients of table 4
    """

    d = 0
    for i, c in enumerate(coeffs):
        d += c * z**i

    return d


def rho_tang(w, coeffs):
    """From Tang 1996, only at 25°C"""
    w = w * 100
    rho0 = 997.1  # density of pure water (at 25°C)
    rho = rho0
    for i, coeff in enumerate(coeffs):
        rho += coeff * w ** (i + 1)
    return rho0, rho



"""Al Ghafri formulae for various salts

Sources
-------
- Al Ghafri et al., Densities of Aqueous MgCl2(aq), CaCl2 (aq), KI(aq),
NaCl(aq), KCl(aq), AlCl3(aq), and (0.964 NaCl + 0.136 KCl)(aq) at
Temperatures Between (283 and 472) K, Pressures up to 68.5 MPa, and
Molalities up to 6 mol·kg -1.
Journal of Chemical & Engineering Data 57, 1288-1304 (2012).
"""

import numpy as np

from ....constants import Tc, Patm
from ...general import SolutionFormula
from ...water.vapor_pressure import VaporPressure_IAPWS
from ...water.density_sat import DensitySat_IAPWS


def rho_alghafri(m, T, P, a, b, c):
    """General formula for density accorging to Al Ghafri 2012.

    Valid for many solutes including NaCl, KCl, CaCl2, etc.

    Inputs
    ------
    m: molality of salt
    T: temperature in K
    a: list of coefficients alpha from table 10
    b: list of coefficients beta from table 9 and 10
    c: list of coefficients gamma from table 9 and 10

    Outputs
    -------
    Density of solution, kg/m^3

    Reference
    ---------
    Al Ghafri et al., Densities of Aqueous MgCl 2 (aq), CaCl 2 (aq), KI(aq),
    NaCl(aq), KCl(aq), AlCl 3 (aq), and (0.964 NaCl + 0.136 KCl)(aq) at
    Temperatures Between (283 and 472) K, Pressures up to 68.5 MPa, and
    Molalities up to 6 mol·kg-1.
    Journal of Chemical & Engineering Data 57, 1288-1304 (2012).

    Notes
    -----
    (from the article's last page)
    These are valid in the temperature range (298.15 to 473.15) K and at
    pressures up to 68.5 MPa for all brines studied except in the case of
    AlCl3 (aq) where the temperature is restricted to the range (298.15 to
    373.15) K. The correlations are valid for all molalities up to
    (5.0,      6.0,       1.06,   6.0,      4.5,     2.0,      and 4.95) mol·kg-1 for
    MgCl2(aq), CaCl2(aq), KI(aq), NaCl(aq), KCl(aq), AlCl3(aq),
    and (0.864 NaCl + 0.136 KCl)(aq), respectively.
    """
    vapor_pressure = VaporPressure_IAPWS()
    p_ref = vapor_pressure.calculate(T)

    density_sat = DensitySat_IAPWS()
    rho_sat = density_sat.calculate(T)

    # reference density (solution density at reference pressure p_ref(T), which
    # is taken to be the vapor pressure of pure water at the given temperature.

    rho_ref = rho_sat

    for i in range(1, 4):
        rho_ref += a[i][0] * m**((i + 1) / 2)  # eq 9

    for i in range(1, 4):
        for j in range(1, 5):
            rho_ref += a[i][j] * m**((i + 1) / 2) * (T / Tc)**((j + 1) / 2)

    # Parameters of the Tammann-Tait equation --------------------------------

    B = 0
    for i in range(2):
        for j in range(4):
            B += b[i][j] * m**i * (T / Tc)**j  # eq10
    B *= 1e6

    C = c[0] + c[1] * m + c[2] * m**(3 / 2)  # eq 11

    # Final calculation ------------------------------------------------------

    rho = rho_ref * (1 - C * np.log((B + P) / (B + p_ref)))**(-1)  # eq 7

    return rho


# ===================== Coefficients for different salts =====================

def coeffs_CaCl2():

    a = np.zeros((4, 5))
    a[1, :] = [2546.760, -39884.946, 102056.957, -98403.334, 33976.048]
    a[2, :] = [-1362.157, 22785.572, -59216.108, 57894.824, -20222.898]
    a[3, :] = [217.778, -3770.645, 9908.135, -9793.484, 3455.587]

    b = np.zeros((2, 4))
    b[0, :] = [-1622.4, 9383.8, -14893.8, 7309.10]
    b[1, :] = [307.24, -1259.10, 2034.03, -1084.94]

    c = np.zeros(3)
    c[:] = [0.11725, -0.00493, 0.00231]

    return a, b, c


def coeffs_KCl():

    a = np.zeros((4, 5))
    a[1, :] = [2332.802, -39637.418, 104801.288, -104266.828, 37030.556]
    a[2, :] = [-1287.572, 23543.994, -63846.097, 65023.561, -23586.370]
    a[3, :] = [206.032, -4003.757, 11128.162, -11595.475, 4295.498]

    b = np.zeros((2, 4))
    b[0, :] = [-1622.4, 9383.8, -14893.8, 7309.10]
    b[1, :] = [211.49, -888.16, 1400.09, -732.79]

    c = np.zeros(3)
    c[:] = [0.11725, -0.00170, 0.00083]

    return a, b, c


def coeffs_KI():

    a = np.zeros((4, 5))
    a[1, :] = [8657.149, -94956.477, 167497.772, -74952.063, -8734.207]
    a[2, :] = [-14420.621, 137360.624, -184940.639, -11953.289, 79847.960]
    a[3, :] = [7340.083, -66939.345, 81446.737, 23983.386, -49031.473]

    b = np.zeros((2, 4))
    b[0, :] = [-1622.4, 9383.8, -14893.8, 7309.10]
    b[1, :] = [241.84, -1030.61, 1548.15, -754.36]

    c = np.zeros(3)
    c[:] = [0.11725, -0.01026, 0.00842]

    return a, b, c


def coeffs_MgCl2():

    a = np.zeros((4, 5))
    a[1, :] = [2385.823, -38428.112, 99526.269, -97041.399, 33841.139]
    a[2, :] = [-1254.938, 21606.295, -56988.274, 56465.943, -19934.064]
    a[3, :] = [192.534, -3480.374, 9345.908, -9408.904, 3364.018]

    b = np.zeros((2, 4))
    b[0, :] = [-1622.4, 9383.8, -14893.8, 7309.10]
    b[1, :] = [358.00, -1597.10, 2609.47, -1383.91]

    c = np.zeros(3)
    c[:] = [0.11725, -0.00789, 0.00142]

    return a, b, c


def coeffs_NaCl():

    a = np.zeros((4, 5))
    a[1, :] = [2863.158, -46844.356, 120760.118, -116867.722, 40285.426]
    a[2, :] = [-2000.028, 34013.518, -88557.123, 86351.784, -29910.216]
    a[3, :] = [413.046, -7125.857, 18640.780, -18244.074, 6335.275]

    b = np.zeros((2, 4))
    b[0, :] = [-1622.4, 9383.8, -14893.8, 7309.10]
    b[1, :] = [241.57, -980.97, 1482.31, -750.98]

    c = np.zeros(3)
    c[:] = [0.11725, -0.00134, 0.00056]

    return a, b, c


# ================== Base class for all al ghafri formulas ===================


class Density_AlGhafri_Base(SolutionFormula):
    """Gathers all info common to all Al Ghafri Formulae"""

    source ='Al Ghafri'

    temperature_unit = 'K'
    temperature_range = (283.15, 473.15)

    concentration_unit = 'm'

    with_water_reference = True

    def calculate(self, m, T):
        rho = rho_alghafri(m, T, Patm, *self.coeffs)
        rho0 = rho_alghafri(0, T, Patm, *self.coeffs)
        return rho0, rho


# ============================ Specific formulas =============================


class Density_CaCl2_AlGhafri_Base(Density_AlGhafri_Base):
    solute = 'CaCl2'
    concentration_range = (0, 6)
    coeffs = coeffs_CaCl2()


class Density_KCl_AlGhafri_Base(Density_AlGhafri_Base):
    solute = 'KCl'
    concentration_range = (0, 4.5)
    coeffs = coeffs_KCl()


class Density_KI_AlGhafri_Base(Density_AlGhafri_Base):
    solute = 'KI'
    concentration_range = (0, 1.05)
    coeffs = coeffs_KI()


class Density_MgCl2_AlGhafri_Base(Density_AlGhafri_Base):
    solute = 'MgCl2'
    concentration_range = (0, 5)
    coeffs = coeffs_MgCl2()


class Density_NaCl_AlGhafri_Base(Density_AlGhafri_Base):
    solute = 'NaCl'
    concentration_range = (0, 6)
    coeffs = coeffs_NaCl()


"""Gathers the formulas for the activity of Na2SO4 solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns one parameter
- a, water activity (dimensionless, range 0-1)
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Clegg et al. : "Thermodynamic Properties of Aqueous Aerosols to High
Supersaturation: II" (1997). Valid at 25°C and for solutions of molality
up to ~17 mol/kg (x ~ 0.23)

- Steiger, M.,
  Crystal growth in porous materials—I:
  The crystallization pressure of large crystals.
  Journal of Crystal Growth 282, 455-469 (2005).
  Valid at 25°C and up to 13.5 mol/kg

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

(some info of domain of validity of expressions in the following paper:)
Dorn, J. & Steiger, M. Measurement and Calculation of Solubilities in the
Ternary System NaCH 3 COO + NaCl + H 2 O from 278 K to 323 K.
J. Chem. Eng. Data 52, 1784-1790 (2007).)

For now, I have assumed validity similar to NaCl (in temperatures)
"""

from .clegg import WaterActivity_Na2SO4_Clegg_Base
from .steiger import WaterActivity_Na2SO4_Steiger2005_Base
from .steiger import WaterActivity_Na2SO4_Steiger2008_Base


class WaterActivity_Na2SO4_Steiger2005(WaterActivity_Na2SO4_Steiger2005_Base):
    """Already defined in steiger module"""
    pass


class WaterActivity_Na2SO4_Steiger2008(WaterActivity_Na2SO4_Steiger2008_Base):
    """Already defined in steiger module"""
    default = True

class WaterActivity_Na2SO4_Clegg(WaterActivity_Na2SO4_Clegg_Base):
    """Already defined in clegg module"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

WaterActivityFormulas_Na2SO4 = (
    WaterActivity_Na2SO4_Steiger2008,
    WaterActivity_Na2SO4_Steiger2005,
    WaterActivity_Na2SO4_Clegg,
)



"""Miscellaneous formatting tools and checks for the aquasol package."""

import functools

import numpy as np


def check_units(units, allowed_units):
    """Check if units are among allowed units, raise exception if not."""
    wrong_units = []
    for unit in units:
        if unit not in allowed_units:
            wrong_units.append(unit)
    if len(wrong_units) > 0:
        raise ValueError(f'{wrong_units} not in allowed units {allowed_units}')


def check_solute(solute, allowed_solutes):
    """Check if solute in allowed solutes to use, raise exception if not"""
    if solute not in allowed_solutes:
        raise ValueError(f"{solute} not in allowed solutes: {allowed_solutes}")


def format_input_type(value):
    """If type of input doesn't support math operations, convert to np array."""
    try:
        value + 1
    except TypeError:
        return np.array(value)
    else:
        return value


def format_output_type(value):
    """Formats the output of inverse functions the same way as the inputs.

    This is because e.g. pynverse outputs an array, but sometimes we want it
    to output a scalar if a scalar was put in.
    """
    try:
        sh = value.shape
    except AttributeError:
        return value   # if not an array, it's a float, so return it rightaway
    else:
        if len(sh) == 0:  # this is to return a scalar if a scalar is used as input
            return value.item()
        else:
            return value


def format_temperature(T, unit_in, unit_out):
    """Format temperature from/to Celsius (C) and Kelvin (K)."""

    allowed_units = 'C', 'K'
    check_units([unit_in, unit_out], allowed_units)

    T = format_input_type(T)  # allows for lists and tuples as inputs

    if unit_in == 'C':
        T_celsius = T
        T_kelvin = T + 273.15
    else:
        T_celsius = T - 273.15
        T_kelvin = T

    if unit_out == 'C':
        return T_celsius
    else:
        return T_kelvin


def format_concentration(concentration, unit_out, solute, converter):
    """Check if concentration unit is ok and convert it to the unit_out unit.

    Parameters
    ----------
    concentration: dict from main function **kwargs (e.g. {'w': 0.1})
    unit_out: the unit to format the value into (e.g. 'w')
    solute: name of the solute (e.g. 'NaCl')
    converter: concentration conversion function (convert or basic_convert)

    Output
    ------
    value in the unit_out unit

    Note
    ----
    Checking if concentration is in the right unit etc. and transforming into
    array if input is tuple, list etc. is done by the converter
    """

    if len(concentration) > 1:
        raise ValueError('concentration must have a single keyword argument')

    if len(concentration) == 0:
        raise ValueError(f'Concentration of {solute} not provided.')

    (unit_in, value), = concentration.items()

    conc = converter(value=value, unit1=unit_in, unit2=unit_out, solute=solute)
    return conc


def make_array(function):
    """Decorator to execute function on arrays even when it's not designed initially to accept arrays"""

    @functools.wraps(function)  # to preserve function signature
    def wrapper(x, *args, **kwargs):
        try:
            iter(x)
        except TypeError:  # Not an array
            return function(x, *args, **kwargs)
        else:
            result = []
            for xi in x:
                y = function(xi, *args, **kwargs)
                result.append(y)
            return np.array(result)
    return wrapper


def make_array_method(method):
    """Decorator to execute method on arrays even when it's not designed initially to accept arrays"""

    @functools.wraps(method)  # to preserve method signature
    def wrapper(self, x, *args, **kwargs):
        try:
            iter(x)
        except TypeError:  # Not an array
            return method(self, x, *args, **kwargs)
        else:
            result = []
            for xi in x:
                y = method(self, xi, *args, **kwargs)
                result.append(y)
            return np.array(result)
    return wrapper


"""Function to calculate the diffusivity in air as a function of temperature

Sources
-------

- Massman, W. J.
  A review of the molecular diffusivities of H2O, CO2, CH4, CO, O3, SO2, NH3,
  N2O, NO, and NO2 in air, O2 and N2 near STP.
  Atmospheric Environment 32, 1111-1127 (1998).

- Marrero, T. R. and Mason E. A.,
  Gaseous diffusion coeffcients.
  Journal of Physics and Chemistry Reference Data 1, 3-118 (1972)
"""

from ..general import WaterFormula


class DiffusivityInAir_Massman(WaterFormula):

    source ='Massman'
    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)
    default = True

    coeffs = {'T0': 273.15}  # K

    def calculate(self,T):
        """Diffusivity of water vapor in air according to Massman 1998

        Input
        -----
        Temperature in K

        Output
        ------
        Diffusivity in m^2 / s

        Reference
        ---------
        Massman, W. J.
        A review of the molecular diffusivities of H2O, CO2, CH4, CO, O3, SO2, NH3,
        N2O, NO, and NO2 in air, O2 and N2 near STP.
        Atmospheric Environment 32, 1111-1127 (1998).

        Notes
        -----
        - Valid between 273.15 K and 373.15 K (0 to 100°C)
        """
        return (0.2178 * (T / self.coeffs['T0']) ** 1.81) * 1e-4


class DiffusivityInAir_MM72(WaterFormula):

    source ='MM72'
    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)

    coeffs = {'T0': 273.15}  # K

    def calculate(self, T):
        """Diffusivity of water vapor in air according to Marrero & Mason (1972)

        (as reported in Massman 1998)

        Input
        -----
        Temperature in K

        Output
        ------
        Diffusivity in m^2 / s

        Reference
        ---------
        Marrero, T. R. and Mason E. A.,
        Gaseous diffusion coeffcients.
        Journal of Physics and Chemistry Reference Data 1, 3-118 (1972)

        Notes
        -----
        - Valid between 273.15 K and 373.15 K (0 to 100°C)
        """
        return (0.209 * (T / self.coeffs['T0']) ** 2.072) * 1e-4


# ========================== WRAP-UP OF FORMULAS =============================

DiffusivityInAirFormulas = (
    DiffusivityInAir_Massman,
    DiffusivityInAir_MM72
)



"""Activity of solutions according to Steiger

NOTE: Almost identical in structure to activity_coefficient.steiger

Sources
-------
- Steiger, M.,
  Crystal growth in porous materials—I:
  The crystallization pressure of large crystals.
  Journal of Crystal Growth 282, 455-469 (2005).
  Valid at 25°C and up to 13.5 mol/kg

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

(some info of domain of validity of expressions in the following paper:)
Dorn, J. & Steiger, M. Measurement and Calculation of Solubilities in the
Ternary System NaCH 3 COO + NaCl + H 2 O from 278 K to 323 K.
J. Chem. Eng. Data 52, 1784-1790 (2007).)
"""

from ...general import SolutionFormula
from ..steiger import coeffs_steiger2005_activity, coeffs_steiger2008_activity
from ..pitzer import PitzerActivity


steiger_coeffs = {
    'Steiger 2005': coeffs_steiger2005_activity,
    'Steiger 2008': coeffs_steiger2008_activity,
}


class WaterActivity_Steiger_Base(SolutionFormula):

    temperature_unit = 'K'
    concentration_unit = 'm'

    with_water_reference = False

    def calculate(self, m, T):
        coeffs = steiger_coeffs[self.source].coeffs(solute=self.solute, T=T)
        pitz = PitzerActivity(T=T, solute=self.solute, **coeffs)
        return pitz.water_activity(m=m)


# =============================== Steiger 2005 ===============================


class WaterActivity_Steiger2005_Base(WaterActivity_Steiger_Base):
    source ='Steiger 2005'
    temperature_range = (298.15, 298.15)


class WaterActivity_NaCl_Steiger2005_Base(WaterActivity_Steiger2005_Base):
    solute = 'NaCl'
    concentration_range = (0, 13.5)


class WaterActivity_Na2SO4_Steiger2005_Base(WaterActivity_Steiger2005_Base):
    solute = 'Na2SO4'
    concentration_range = (0, 12)


# =============================== Steiger 2008 ===============================


class WaterActivity_Steiger2008_Base(WaterActivity_Steiger_Base):
    source ='Steiger 2008'
    temperature_range = (273.15, 323.15)


class WaterActivity_NaCl_Steiger2008_Base(WaterActivity_Steiger2008_Base):
    solute = 'NaCl'
    concentration_range = (0, 15)


class WaterActivity_Na2SO4_Steiger2008_Base(WaterActivity_Steiger2008_Base):
    solute = 'Na2SO4'
    concentration_range = (0, 12)


class WaterActivity_KCl_Steiger2008_Base(WaterActivity_Steiger2008_Base):
    solute = 'KCl'
    concentration_range = (0, 15)  # NOT SURE (see above)



"""Properties of solutions."""

# TODO: add viscosity, diffusivity
# TODO: add saturation concentration of different solutes
# TODO: add partial molar volumes

from .properties import activity_coefficient, water_activity
from .properties import density, surface_tension
from .properties import refractive_index, electrical_conductivity
from .properties import solubility

from .convert import convert
from .invert import aw_to_conc
from .extend import osmotic_pressure, osmotic_coefficient, aw_saturated

from ..formulas.solutions.ionic import ion_quantities, ionic_strength



"""Properties of pure water."""

from ..properties import WaterProperty

from ..formulas.water.density_atm import DensityAtmFormulas
from ..formulas.water.density_sat import DensitySatFormulas
from ..formulas.water.diffusivity_in_air import DiffusivityInAirFormulas
from ..formulas.water.surface_tension import SurfaceTensionFormulas
from ..formulas.water.vapor_pressure import VaporPressureFormulas
from ..formulas.water.viscosity_atm import ViscosityAtmFormulas


class DensityAtm(WaterProperty):
    """Density of ambient pure water as a function of temperature [kg/m^3].

    Examples
    --------
    >>> from aquasol.water import density_sat as rho
    >>> rho()  # returns the density of water (rho) at 25°C
    >>> rho(20)                  # rho  at 20°C
    >>> rho([0, 10, 20, 30])     # rho at various temperatures in Celsius
    >>> rho(300, 'K')            # rho at 300K
    """

    quantity = 'density (sat.)'
    unit = '[kg/m^3]'

    Formulas = DensityAtmFormulas


class DensitySat(WaterProperty):
    """Density of saturated liquid water as a function of temperature [kg/m^3].

    Examples
    --------
    >>> from aquasol.water import density_sat as rho
    >>> rho()  # returns the denisty of water (rho) at 25°C
    >>> rho(20)                  # rho  at 20°C
    >>> rho([0, 10, 20, 30])     # rho at various temperatures in Celsius
    >>> rho(300, 'K')            # rho at 300K
    """

    quantity = 'density (sat.)'
    unit = '[kg/m^3]'

    Formulas = DensitySatFormulas


class DiffusivityInAir(WaterProperty):
    """Diffusivity of water vapor as a function of temperature [m^2/s].

    Examples
    --------
    >>> from aquasol.water import diffusivity_in_air as d
    >>> d()  # returns the diffusivity of water in air at 25°C
    >>> d(20)                  # at 20°C
    >>> d([0, 10, 20, 30])     # at various temperatures in Celsius
    >>> d(300, 'K')            # at 300K
    """

    quantity = 'vapor diffusivity in air'
    unit = '[m^2/s]'

    Formulas = DiffusivityInAirFormulas


class SurfaceTension(WaterProperty):
    """Surface tension of pure water as a function of temperature [N/m]

    Examples
    --------
    >>> from aquasol.water import surface_tension as sigma
    >>> sigma()  # returns the surface tension of water (sigma) at 25°C
    >>> sigma(20)                  # sigma  at 20°C
    >>> sigma([0, 10, 20, 30])     # sigma at various temperatures in Celsius
    >>> sigma(300, 'K')            # sigma at 300K
    """

    quantity = 'surface tension'
    unit = '[N/m]'

    Formulas = SurfaceTensionFormulas


class VaporPressure(WaterProperty):
    """Saturation vapor pressure of water as a function of temperature [Pa]

    Examples
    --------
    >>> from aquasol.water import vapor_pressure as psat
    >>> psat()  # returns the saturation vapor pressure of water at 25°C
    >>> psat(20)                   # at 20°C
    >>> psat([0, 10, 20, 30])      # at various temperatures in Celsius
    >>> psat(300, 'K')             # at 300K
    >>> psat(15, source='Wexler')  # at 15°C using Wexler equation
    """

    quantity = 'saturated vapor pressure'
    unit = '[Pa]'

    Formulas = VaporPressureFormulas


class ViscosityAtm(WaterProperty):
    """Viscosity of water at ambient pressure as a function of temperature [Pa.s]

    Examples
    --------
    >>> from aquasol.water import viscosity_atm as mu
    >>> mu()  # returns the diffusivity of water in air at 25°C
    >>> mu(20)                  # at 20°C
    >>> mu([0, 10, 20, 30])     # at various temperatures in Celsius
    >>> mu(300, 'K')            # at 300K
    """

    quantity = 'viscosity (atm.)'
    unit = '[Pa.s]'

    Formulas = ViscosityAtmFormulas


density_atm = DensityAtm()
density_sat = DensitySat()
diffusivity_in_air = DiffusivityInAir()
surface_tension = SurfaceTension()
vapor_pressure = VaporPressure()
viscosity_atm = ViscosityAtm()



"""Gathers the formulas for the activity coefficients of KCl solutions.

Sources
-------

- Steiger, M., Kiekbusch, J. & Nicolai,
  An improved model incorporating Pitzer's equations for calculation of
  thermodynamic properties of pore solutions implemented into an efficient
  program code.
  Construction and Building Materials 22, 1841-1850 (2008).

NOTE: I could not find explicit info about validity domain for the KCl
      formulas in Steiger, so I kept ~ same values as for NaCl
"""

from .steiger import ActivityCoefficient_KCl_Steiger2008_Base
from .tang import ActivityCoefficient_KCl_Tang_Base


class ActivityCoefficient_KCl_Steiger2008(ActivityCoefficient_KCl_Steiger2008_Base):
    """Already defined in steiger module"""
    default = True


class ActivityCoefficient_KCl_Tang(ActivityCoefficient_KCl_Tang_Base):
    """Already defined in tang module"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

ActivityCoefficientFormulas_KCl = (
    ActivityCoefficient_KCl_Steiger2008,
    ActivityCoefficient_KCl_Tang,
)



from aquasol.water import vapor_pressure, surface_tension
from aquasol.water import density_sat, density_atm, molar_volume
from aquasol.water import diffusivity_in_air, viscosity_atm
from aquasol.water import dewpoint, kelvin_humidity, kelvin_radius, kelvin_pressure
import numpy as np

# =========================== Test Vapor Pressure ============================

def test_psat_1():
    p = vapor_pressure(1)
    assert round(p) == 657

def test_psat_2():
    p1 = vapor_pressure(T=1)
    p2 = vapor_pressure(274.15, 'K')
    assert p1 == p2

def test_psat_3():
    p = vapor_pressure()
    assert round(p) == 3170

def test_psat_4():
    p = vapor_pressure(T=293, unit='K')
    assert type(p) is float

def test_psat_5():
    p = vapor_pressure(T=[5, 15, 25])
    assert type(p) is np.ndarray

def test_psat_6():
    p = vapor_pressure(T=[5, 15, 25])
    assert p.shape == (3,)

def test_psat_7():
    p = vapor_pressure(T=[5, 15, 25])
    assert round(p[1]) == 1706

# =========================== Test Surface Tension ===========================

def test_sigma_1():
    s = surface_tension(1)
    assert round(s, 4) == 0.0755

def test_sigma_2():
    s1 = surface_tension(T=1)
    s2 = surface_tension(274.15, 'K')
    assert s1 == s2

def test_sigma_3():
    s = surface_tension()
    assert round(s, 3) == 0.072

def test_sigma_4():
    s = surface_tension(T=293, unit='K')
    assert type(s) is float

def test_sigma_5():
    s = surface_tension(T=[0, 60, 100])
    assert type(s) is np.ndarray

def test_sigma_6():
    s = surface_tension(T=[0, 60, 100])
    assert s.shape == (3,)

def test_sigma_7():
    s = surface_tension(T=[0, 60, 100])
    assert round(s[1], 3) == 0.066

# ========================= Test Atmospheric Density =========================

def test_rhoatm_1():
    rho = density_atm(4)
    assert round(rho) == 1000

def test_rhoatm_2():
    s1 = density_atm(T=1)
    s2 = density_atm(274.15, 'K')
    assert s1 == s2

def test_rhoatm_3():
    rho = density_atm()
    assert round(rho) == 997

def test_rhoatm_4():
    rho = density_atm(T=293, unit='K')
    assert type(rho) is float

def test_rhoatm_5():
    rho = density_atm(T=[0, 60, 100])
    assert type(rho) is np.ndarray

def test_rhoatm_6():
    rho = density_atm(T=[0, 60, 100])
    assert rho.shape == (3,)

def test_rhoatm_7():
    rho = density_atm(T=[0, 60, 100])
    assert round(rho[1]) == 983

# ========================== Test Saturated Density ==========================

def test_rhosat_1():
    rho = density_sat(4)
    assert round(rho) == 1000

def test_rhosat_2():
    s1 = density_sat(T=1)
    s2 = density_sat(274.15, 'K')
    assert s1 == s2

def test_rhosat_3():
    rho = density_sat()
    assert round(rho) == 997

def test_rhosat_4():
    rho = density_sat(T=293, unit='K')
    assert type(rho) is float

def test_rhosat_5():
    rho = density_sat(T=[0, 60, 100])
    assert type(rho) is np.ndarray

def test_rhosat_6():
    rho = density_sat(T=[0, 60, 100])
    assert rho.shape == (3,)

def test_rhosat_7():
    rho = density_sat(T=[0, 60, 100])
    assert round(rho[1]) == 983


# ============================= Test molar volume ============================


def test_vm_1():
    vm = molar_volume()
    assert round(vm * 1e5, 3) == 1.807

def test_vm_2():
    vm = molar_volume(condition='atm')
    assert round(vm * 1e5, 3) == 1.807

def test_vm_3():
    vm = molar_volume(T=20)
    assert round(vm * 1e5, 3) == 1.805

def test_vm_4():
    vms = molar_volume(T=[15, 25])
    assert round(vms[1] * 1e5, 3) == 1.807


# ========================= Test Diffusivity in air ==========================

def test_diffusivity_1():
    d = diffusivity_in_air()
    assert round(d * 1e5, 2) == 2.55

def test_diffusivity_2():
    d = diffusivity_in_air(source='MM72', T=50)
    assert round(d * 1e5, 2) == 2.96

# ========================= Test Viscosity (Ambient) =========================

def test_viscosity_1():
    mu = viscosity_atm()
    assert round(mu * 1e3, 2) == 0.89

def test_viscosity_2():
    mu = viscosity_atm(T=100)
    assert round(mu * 1e3, 2) == 0.28

# ============================== Test Dewpoint ===============================

def test_dewpoint_p_1():
    dp = dewpoint(p=1000)  # Dew point of a vapor at 1kPa
    assert round(dp, 1) == 7

def test_dewpoint_p_2():
    dp = dewpoint(p=1000, unit='K')  # Same, but temperature is returned in K
    assert round(dp, 1) == 280.1

def test_dewpoint_p_3():
    dp = dewpoint('K', p=1000)  # same thing
    assert round(dp, 1) == 280.1

def test_dewpoint_p_4():
    """Check that return float when input is single value."""
    dp = dewpoint(p=1000)
    assert type(dp) is float

def test_dewpoint_p_5():
    """Check that array returned when input is a list."""
    dp = dewpoint(p=[1000, 2000, 3000])
    assert type(dp) is np.ndarray

def test_dewpoint_p_6():
    """Check that length of array returned corresponds to input"""
    dp = dewpoint(p=[1000, 2000, 3000])
    assert dp.shape == (3,)

def test_dewpoint_p_7():
    """Check that values in array are ok."""
    dp = dewpoint(p=[1000, 2000, 3000])
    assert round(dp[2], 1) == 24.1

def test_dewpoint_rhaw_1():
    dp = dewpoint(rh=50)  # Dew point at 50%RH and 25°C (default)
    assert round(dp, 1) == 13.9

def test_dewpoint_rhaw_2():
    dp1 = dewpoint(rh=50)  # Dew point at 50%RH and 25°C (default)
    dp2 = dewpoint(aw=0.5)  # same thing
    assert dp1 == dp2

def test_dewpoint_rhaw_3():
    dp = dewpoint(aw=0.5, T=20)  # same thing, but at 20°C
    assert round(dp, 1) == 9.3

def test_dewpoint_rhaw_4():
    dp = dewpoint('K', 300, aw=0.5)  # same thing, but at 300K (dewpoint also in K)
    assert round(dp, 1) == 288.7

def test_dewpoint_rhaw_5():
    dp = dewpoint(rh=50)  # Single input should result in float output
    assert type(dp) is float

def test_dewpoint_rhaw_6():
    dp = dewpoint(aw=[0.3, 0.5, 0.7, 0.9])  # It is possible to input lists, tuples, arrays
    assert type(dp) is np.ndarray

def test_dewpoint_rhaw_7():
    dp = dewpoint(aw=[0.3, 0.5, 0.7, 0.9])
    assert dp.shape == (4,)

def test_dewpoint_rhaw_8():
    dp = dewpoint(aw=[0.3, 0.5, 0.7, 0.9])
    assert round(dp[3], 1) == 23.2

# ========================== Test Kelvin relations ===========================

# Pressure -------------------------------------------------------------------

def test_kelvin_pressure_1():
    P = - kelvin_pressure(aw=0.8) / 1e6  # Kelvin pressure in MPa at 80%RH and T=25°C
    assert round(P) == 31

def test_kelvin_pressure_2():
    P = - kelvin_pressure(rh=80) / 1e6           # same
    assert round(P) == 31

def test_kelvin_pressure_3():
    P = - kelvin_pressure(p=1000, T=293.15, unit='K') / 1e6     # at 1000Pa, 20°C
    assert round(P) == 115

def test_kelvin_pressure_4():
    P = - kelvin_pressure(aw=[0.5, 0.7, 0.9]) / 1e6    # possible to use iterables
    assert round(P[1]) == 49

# Radius ---------------------------------------------------------------------
def test_kelvin_radius_1():
    r = kelvin_radius(aw=0.8) * 1e9  # Kelvin radius at 80%RH and T=25°C
    assert round(r, 1) == 4.7

def test_kelvin_radius_2():
    r = kelvin_radius(rh=80) * 1e9           # same
    assert round(r, 1) == 4.7

def test_kelvin_radius_3():
    r = kelvin_radius(rh=80, ncurv=1) * 1e9   # assume cylindrical meniscus instead of spherical
    assert round(r, 2) == 2.35

def test_kelvin_radius_4():
    r = kelvin_radius(p=1000, T=20) * 1e9      # at 1000Pa, 20°C
    assert round(r, 2) == 1.27

def test_kelvin_radius_5():
    r = kelvin_radius(p=1000, T=293.15, unit='K') * 1e9      # same
    assert round(r, 2) == 1.27

def test_kelvin_radius_6():
    r = kelvin_radius(aw=[0.5, 0.7, 0.9]) * 1e9    # possible to use iterables
    assert round(r[1], 2) == 2.94

# Humidity ---------------------------------------------------------------------

def test_kelvin_humidity_1():
    a = kelvin_humidity(r=4.7e-9)  # activity corresponding to Kelvin radius of 4.7 nm at 25°C
    assert round(a, 2) == 0.80

def test_kelvin_humidity_2():
    rh = kelvin_humidity(r=4.7e-9, out='rh')  # same, but expressed in %RH instead of activity
    assert round(rh) == 80

def test_kelvin_humidity_3():
    p = kelvin_humidity(r=4.7e-9, out='p')  # same, but in terms of pressure (Pa)
    assert round(p) == 2536

def test_kelvin_humidity_4():
    p = kelvin_humidity(r=4.7e-9, out='p', T=293.15, unit='K')  # at a different temperature
    assert round(p) == 1860

def test_kelvin_humidity_5():
    a = kelvin_humidity(r=4.7e-9, ncurv=1)  # cylindrical interface
    assert round(a, 2) == 0.89

def test_kelvin_humidity_6():
    aa = kelvin_humidity(r=[3e-9, 5e-9])  # with iterables
    assert round(aa[1], 2) == 0.81

def test_kelvin_humidity_7():     # with liquid pressure as input instead of r
    a = kelvin_humidity(P=-30e6)
    assert round(a, 2) == 0.80

def test_kelvin_humidity_8():   # liquid pressure as input, vapor pressure out
    p = kelvin_humidity(P=-30e6, out='p', T=293.15, unit='K')
    assert round(p) == 1873

def test_kelvin_humidity_9():   # iterable liquid pressure as input
    aa = kelvin_humidity(P=[-30e6, -50e6])
    assert round(aa[1], 2) == 0.69



"""Gathers the formulas for the density of CaCl2 solutions.

Note
----
When adding source, make sure to make a function that has two parameters:
- w (weight fraction), range 0-1 or other concentration quantity
- T (temperature), in K
and returns two parameters:
- rho0, density of pure water in kg / m^3
- rho, density of solution in kg / m^3
Also, add the name of the function to the formulas dictionary at the end of the
file.

Sources
-------
- Al Ghafri et al., Densities of Aqueous MgCl2(aq), CaCl2 (aq), CaCl2(aq),
NaCl(aq), KCl(aq), AlCl3(aq), and (0.964 NaCl + 0.136 KCl)(aq) at
Temperatures Between (283 and 472) K, Pressures up to 68.5 MPa, and
Molalities up to 6 mol·kg -1.
Journal of Chemical & Engineering Data 57, 1288-1304 (2012).

- Conde, M. R., Properties of aqueous solutions of lithium and calcium
chlorides: formulations for use in air conditioning equipment design.
International Journal of Thermal Sciences 43, 367-382 (2004).

- Krumgalz, B. S., Pogorelsky, R. & Pitzer, K. S.
  Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation
  Concentration at 298.15 °K Represented by Pitzer's Ion-Interaction Equations.
  Journal of Physical and Chemical Reference Data 25, 663-689 (1996).
"""

import numpy as np

from ...general import SolutionFormula
from ...water.density_atm import DensityAtm_IAPWS

from .misc import relative_rho_conde
from .al_ghafri import Density_CaCl2_AlGhafri_Base
from .krumgalz import Density_CaCl2_Krumgalz_Base


class Density_CaCl2_Conde(SolutionFormula):

    source ='Conde'
    solute = 'CaCl2'

    temperature_unit = 'K'
    temperature_range = (273.15, 373.15)

    concentration_unit = 'r'
    concentration_range = (0, 1.5)

    default = True
    with_water_reference = True

    coeffs = 1, 0.836014, -0.436300, 0.105642

    def calculate(self, z, T):
        d = relative_rho_conde(z, self.coeffs)
        density_atm = DensityAtm_IAPWS()
        rho0 = density_atm.calculate(T=T)
        return rho0, rho0 * d


class Density_CaCl2_AlGhafri(Density_CaCl2_AlGhafri_Base):
    """Already defined in Al Ghafri module"""
    default = True

class Density_CaCl2_Krumgalz(Density_CaCl2_Krumgalz_Base):
    """Already defined in Krumgalz module and not default here"""
    pass


# ========================== WRAP-UP OF FORMULAS =============================

Density_CaCl2_Formulas = (
    Density_CaCl2_Conde,
    Density_CaCl2_AlGhafri,
    Density_CaCl2_Krumgalz,
)


"""Invert some properties for water (e.g., dewpoint from vapor_pressure)."""

# Non-standard imports
from pynverse import inversefunc

# Local imports
from . import vapor_pressure

from ..format import format_temperature, format_output_type
from ..humidity import format_humidity


def dewpoint(unit='C', T=None, source=None, **humidity):
    """Inverts vapor_pressure() to calculate dew point at a given humidity.

    Inputs
    ------
    - unit: temperature unit of dewpoint, can be 'C' or 'K' (default 'C')
    - T: system temperature, required only if rh or aw are used as humidity
    input value, but optional if p is used. Default None, i.e 25°C.
    - source: literature source for the calculation of vapor pressure
              (default: None, i.e. Auto); see water.vapor_pressure
    - humidity kwargs: can be 'rh=' (relative humidity in %), 'aw=' (vapor
    activity = rh / 100), 'p=' (partial water vapor pressure).

    Output
    ------
    Dewpoint Temperature

    Examples
    --------
    >>> dewpoint(p=1000)  # Dew point of a vapor at 1kPa
    6.970481357025221
    >>> dewpoint(p=1000, unit='K')  # Same, but temperature is returned in K
    280.1204813570252
    >>> dewpoint('K', p=1000)  # same thing
    280.1204813570252
    >>> dewpoint(rh=50)  # Dew point at 50%RH and 25°C (default)
    13.864985413550704
    >>> dewpoint(aw=0.5)  # same thing
    13.864985413550704
    >>> dewpoint(aw=0.5, T=20)  # same thing, but at 20°C
    9.273546905501904
    >>> dewpoint('K', 300, aw=0.5)  # same thing, but at 300K (dewpoint also in K)
    288.71154892380787
    >>> dewpoint(aw=[0.5, 0.7])  # It is possible to input lists, tuples, arrays
    array([ 9.27354606, 14.36765209])
    """
    p = format_humidity(
        unit=unit,
        T=T,
        source=source,
        out='p',
        **humidity,
    )
    source = vapor_pressure.get_source(source=source)
    trange_source = vapor_pressure.formulas[source].temperature_range
    tunit_source = vapor_pressure.formulas[source].temperature_unit

    # invert vapor pressure function to get dewpoint -------------------------

    def psat(Ts):
        return vapor_pressure(Ts, unit=tunit_source, source=source)

    dewpoint_calc = inversefunc(psat, domain=trange_source)

    try:
        dpt = dewpoint_calc(p)
        T_out = format_temperature(T=dpt, unit_in=tunit_source, unit_out=unit)
        return format_output_type(T_out)

    except ValueError:
        msg = "Error, probably because T outside of Psat formula validity range"
        raise ValueError(msg)



"""Dutcher formula for the surface tension of NaCl solutions.

Source
------
- Dutcher: Dutcher, C. S., Wexler, A. S. & Clegg, S. L.
  Surface Tensions of Inorganic Multicomponent Aqueous Electrolyte Solutions and Melts.
  J. Phys. Chem. A 114, 12216-12230 (2010).
"""

import numpy as np

from ...general import SolutionFormula
from .misc import sigma_iapws


def sigma_dutcher(x, T, coeffs_table3, coeffs_table5):
    """General formula for surface tension accorging to Dutcher 2010.

    Valid for many solutes including NaCl, KCl, HCl, Na2S04, CaCl2, etc.

    Inputs
    ------
    x: mole fraction of salt
    T: temperature in K
    coeffs_table3: tuple or list of coefficients c1, c2 from table 3
    coeffs_table5: tuple or list of coefficients aws, bws, asw, bsw (table 5)

    Outputs
    -------
    Surface tension of solution, N/m

    Notes
    -----
    - Validity range for temperature and concentration are given in Table 2.
    - The original paper has a value for the critical point of water slightly
    off (647.15 instead of 647.096), but the difference is not noticeable.
    This value is used in the calculation of the surface tension of pure water
    from the IAPWS formula.

    Reference
    ---------
    Dutcher: Dutcher, C. S., Wexler, A. S. & Clegg, S. L. Surface Tensions of
    Inorganic Multicomponent Aqueous Electrolyte Solutions and Melts.
    J. Phys. Chem. A 114, 12216-12230 (2010).
    """

    # Coefficients of Table 3
    c1, c2 = coeffs_table3
    # Coefficients of Table 5
    aws, bws, asw, bsw = coeffs_table5

    xw = 1 - x  # mole fraction of water

    # Surface tension of water (Eq. 10)
    gw = sigma_iapws(T) * 1e3    # in mN / m

    # Surface tension of molten salt (Eq. 12)
    gs = c1 + c2 * T    # in mN / m

    # Surface tension of solution (Eq. 5)
    Fws = aws + bws * T
    Fsw = asw + bsw * T
    gna = np.exp(xw * np.log(gw + Fws * x) + x * np.log(gs + Fsw * xw))  # mN / m

    return gna * 1e-3


def coeffs_CaCl2():

    # Coefficients (Table 3)
    c1 = 195.67     # note - other values possible: (189, -0.03952)
    c2 = -0.04541
    # Coefficients (Table 5)
    aws = -19.766
    bws = 0.575
    asw = 0
    bsw = 0

    coeffs_table3 = c1, c2
    coeffs_table5 = aws, bws, asw, bsw

    return coeffs_table3, coeffs_table5


def coeffs_KCl():

    # Coefficients (Table 3)
    c1 = 177.61
    c2 = -0.07519
    # Coefficients (Table 5)
    aws = -117.33
    bws = 0.489
    asw = 0
    bsw = 0

    coeffs_table3 = c1, c2
    coeffs_table5 = aws, bws, asw, bsw

    return coeffs_table3, coeffs_table5


def coeffs_MgCl2():

    # Coefficients (Table 3)
    c1 = 65.343
    c2 = -0.003073
    # Coefficients (Table 5)
    aws = 1069.9
    bws = -2.86
    asw = 0
    bsw = 0

    coeffs_table3 = c1, c2
    coeffs_table5 = aws, bws, asw, bsw

    return coeffs_table3, coeffs_table5


def coeffs_Na2SO4():

    # Coefficients (Table 3)
    c1 = 269
    c2 = -0.066
    # Coefficients (Table 5)
    aws = 126
    bws = 0
    asw = -194.45
    bsw = 0

    coeffs_table3 = c1, c2
    coeffs_table5 = aws, bws, asw, bsw

    return coeffs_table3, coeffs_table5


def coeffs_NaCl():

    # Coefficients (Table 3)
    c1 = 191.16     # note - other values possible: (193.48, -0.07188)
    c2 = -0.0747
    # Coefficients (Table 5)
    aws = 232.54
    bws = -0.245
    asw = -142.42
    bsw = 0

    coeffs_table3 = c1, c2
    coeffs_table5 = aws, bws, asw, bsw

    return coeffs_table3, coeffs_table5


class SurfaceTension_Dutcher_Base(SolutionFormula):

    source ='Dutcher'

    temperature_unit = 'K'
    concentration_unit = 'x'

    with_water_reference = True

    def calculate(self, x, T):
        """Surface tension calculated from Dutcher 2010.
        Input: mole fraction x, temperature T in K."""
        sigma_w = sigma_iapws(T)
        sigma = sigma_dutcher(x, T, *self.coeffs)

        return sigma_w, sigma


class SufaceTension_CaCl2_Dutcher_Base(SurfaceTension_Dutcher_Base):
    solute = 'CaCl2'
    temperature_range = (243.15, 373.15)
    concentration_range = (0, 0.117)
    coeffs = coeffs_CaCl2()


class SufaceTension_KCl_Dutcher_Base(SurfaceTension_Dutcher_Base):
    solute = 'KCl'
    temperature_range = (265.15, 353.15)
    concentration_range = (0, 0.138)  # Estimated from m_max = 8.86
    coeffs = coeffs_KCl()


class SufaceTension_MgCl2_Dutcher_Base(SurfaceTension_Dutcher_Base):
    solute = 'MgCl2'
    temperature_range = (283.15, 343.15)
    concentration_range = (0, 0.0944)
    coeffs = coeffs_MgCl2()


class SufaceTension_Na2SO4_Dutcher_Base(SurfaceTension_Dutcher_Base):
    solute = 'Na2SO4'
    temperature_range = (273.15, 466.55)
    concentration_range = (0, 0.0446)
    coeffs = coeffs_Na2SO4()


class SufaceTension_NaCl_Dutcher_Base(SurfaceTension_Dutcher_Base):
    solute = 'NaCl'
    temperature_range = (263.13, 473.15)
    concentration_range = (0, 0.145)
    coeffs = coeffs_NaCl()



"""Thermodynamic & physico-chemical properties for water and aqueous solutions

Copyright Olivier Vincent (2020-2024)
(ovinc.py@gmail.com)

This software is a computer program whose purpose is to provide the
properties of water and aqueous solutions as a function of temperature
and/or concentration (along with other useful tools).

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software. You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author, the holder of the
economic rights, and the successive licensors have only limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading, using, modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean that it is complicated to manipulate, and that also
therefore means that it is reserved for developers and experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and, more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
"""

# TODO: Switch automatically to another equation if outside of range?


# Shortcuts ------------------------------------------------------------------

from .water import vapor_pressure as ps
from .water import dewpoint as dp
from .water import kelvin_pressure as kp
from .water import kelvin_humidity as kh
from .water import kelvin_radius as kr
from .water import molar_volume as vm

from .solutions import water_activity as aw
from .solutions import aw_to_conc as ac
from .solutions import convert as cv

# ----------------------------------------------------------------------------

from importlib_metadata import version

__author__ = 'Olivier Vincent'
__version__ = version('aquasol')
__license__ = 'CeCILL-2.1'



"""Properties of pure water."""

from .properties import density_atm, density_sat, diffusivity_in_air
from .properties import surface_tension, vapor_pressure, viscosity_atm

from .invert import dewpoint
from .extend import kelvin_humidity, kelvin_radius, kelvin_pressure
from .extend import molar_volume



"""Convert to/from various parameters (molarity, molality, x, w, etc.)

SOURCES
-------
Personal calculations, and:
ionic strength expressed as molar fraction is used in Clegg et al. AST, 1997
ionic strength expressed as molality is more common, e.g. Pitzer 1973
"""

# Standard Library
import warnings

# Non-standard imports
from pynverse import inversefunc
import numpy as np

# Local imports
from ..constants import solute_list
from ..constants import molar_mass
from ..format import check_solute, check_units
from ..format import format_input_type, format_output_type, format_concentration
from ..properties import SolutionProperty

from ..formulas.solutions.basic_conversions import basic_convert
from ..formulas.solutions.basic_conversions import allowed_units as basic_units

from ..formulas.solutions.density import DensityFormulas


# ================================== Config ==================================


add_units = ['c']
allowed_units = basic_units + add_units


# ============== Reduced density class (avoid circular imports) ==============


class Density_Basic(SolutionProperty):
    """Density property with partial converter (without molarity).

    Is used to prevent circular import problems
    (because SolutionProperty is also used to define the density function used
    in convert())
    """
    # See SolutionProperty for explantion of staticmethod()
    converter = staticmethod(basic_convert)
    Formulas = DensityFormulas
    quantity = 'density'
    unit = ['kg/m^3']


density_basic = Density_Basic()


# ============================= MOLARITY FUNCTIONS ===========================


def w_to_molarity(w, solute, T=25, unit='C', source=None):
    """Calculate molarity of solute from weight fraction at temperature T in °C"""
    check_solute(solute, solute_list)
    M = molar_mass(solute)
    rho = density_basic(solute=solute, T=T, unit=unit, source=source, w=w)
    return rho * w / M


def _get_max_w(solute, T=25, unit='C', source=None, wmin=0, wmax=0.999):
    """Detect if density has a maximum, in which case the inversion fails."""
    ww = np.linspace(wmin, wmax, num=200)
    with warnings.catch_warnings():      # this is to avoid always warnings
        warnings.simplefilter('ignore')  # which pop up due to wmax being high
        rho = density_basic(solute=solute, T=T, unit=unit, source=source, w=ww)
    imax = np.argmax(rho)
    return ww[imax] # will be equal to wmax if function only increases


def molarity_to_w(c, solute, T=25, unit='C', source=None, wmin=0, wmax=0.999):
    """Calculate weight fraction of solute from molarity at temperature T in °C.

    Note: can be slow because of inverting the function each time.
    """
    check_solute(solute, solute_list)

    def molarity(w):
        with warnings.catch_warnings():      # this is to avoid always warnings
            warnings.simplefilter('ignore')  # which pop up due to wmax being high
            return w_to_molarity(
                w=w,
                solute=solute,
                T=T,
                unit=unit,
                source=source,
            )

    # replace wmax if density goes through a maximum
    # (in order to prevent function inversion to fail)
    wmax = _get_max_w(
        solute=solute,
        T=T,
        unit=unit,
        source=source,
        wmin=wmin,
        wmax=wmax,
    )

    # Also prevents inversion fail, and provides more explicit error msg -----
    cmax = molarity(wmax)
    # Line below should always work because c is usually converted to np array
    # if it is a list or tuple etc.
    test = (c > cmax)

    try:  # won't work for an array
        cmax_exceeded = bool(test)
    except ValueError:
        cc = np.array(c).flatten()
        cmax_exceeded = any(cc > cmax)

    if cmax_exceeded:
        src = density_basic.get_source(solute=solute, source=source)
        msg = (
            f"Requested molality (c={c}) exceeds maximum available with "
            f"{src}'s density formula (cmax={round(cmax):_}), "
            f"corresponding to wmax={wmax:.3f}"
        )
        raise ValueError(msg)
    # ------------------------------------------------------------------------

    weight_fraction = inversefunc(molarity, domain=[wmin, wmax])
    w = weight_fraction(c)

    # This is to give a warning if some value(s) of w out of range when
    formula = density_basic.get_formula(solute=solute, source=source)
    c = format_concentration(
            concentration={'w': w},
            unit_out=formula.concentration_unit,
            solute=solute,
            converter=density_basic.converter,
        )
    formula.check_validity_range('concentration', value=c)

    return format_output_type(w)


# =========================== MAIN CONVERT FUNCTION ==========================


def convert(
    value,
    unit1,
    unit2,
    solute='NaCl',
    T=25,
    unit='C',
    density_source=None,
    density_wmin=0,
    density_wmax=0.999,
):
    """Convert between different concentration units for solutions.

    Parameters
    ----------
    - value (float): value to convert
    - unit1 (str): its unit.
    - unit2 (str): unit to convert to.
    - solute (str): name of solute (default 'NaCl').
    - T: temperature
    - unit: unit of temperature (should be 'C' or 'K'), only used for molarity

    Additional parameters are available when converting to/from molarity,
    because knowledge of solution density is required:
    - density_source: which formula to use to calculate density when converting
                      to/from molarity (None = default).
    - density_wmin: min mass fraction to consider when inverting molarity(w)
                    for iterative search (only when converting FROM molarity)
    - density_wmin: max mass fraction to consider when inverting molarity(w)
                    for iterative search (only when converting FROM molarity)

    solute has to be in the solute list in the constants module and in the
    solutes with density data if unit1 or unit2 is molarity ('c').

    unit1 and unit2 have to be in the allowed units list :
    'x' (mole fraction), 'w' (weight fraction), 'm' (molality), 'r'
    (ratio of mass of solute to mass of solvent), 'c' (molarity in mol/m^3)

    Output
    ------
    Converted value (dimensionless or SI units)

    Examples
    --------
    - convert(0.4, 'w', 'x'): mass fraction of 0.4 into mole fraction for NaCl
    - convert(10, 'm', 'w'): molality of 10 mol/kg into mass fraction for NaCl
    - convert(10, 'm', 'w', 'LiCl'): same but for LiCl.
    - convert(5000, 'c', 'x'): molarity of 5 mol/m^3 to mole fraction for NaCl
    (assuming a temperature of T=25°C)
    - convert(5000, 'c', 'x', T=30): same but at 30°C.
    - convert(5000, 'c', 'x', T=293, unit='K'): same but at 293K
    - convert(5000, 'c', 'x', solute='LiCl'): for LiCl at 25°C
    """
    units = [unit1, unit2]
    check_units(units, allowed_units)

    value = format_input_type(value)  # allows for lists and tuples as inputs

    # No need to calculate anything if the in and out units are the same -----
    if unit1 == unit2:
        return value

    check_solute(solute, solute_list)

    if unit1 in basic_units and unit2 in basic_units:
        return basic_convert(
            value=value,
            unit1=unit1,
            unit2=unit2,
            solute=solute,
        )

    # Check if it's unit1 which is a "fancy" unit and convert to w if so.
    if unit1 == 'c':
        w = molarity_to_w(
            c=value,
            solute=solute,
            T=T,
            unit=unit,
            source=density_source,
            wmin=density_wmin,
            wmax=density_wmax,
        )
        value_in = w
        unit_in = 'w'
    else:
        value_in = value
        unit_in = unit1

    if unit2 in basic_units:   # If unit2 is basic, the job is now easy
        return basic_convert(
            value=value_in,
            unit1=unit_in,
            unit2=unit2,
            solute=solute,
        )

    else:  # If not, first convert to w, then again to the asked unit

        w = basic_convert(value_in, unit_in, 'w', solute)
        if unit2 == 'c':
            return w_to_molarity(
                w=w,
                solute=solute,
                T=T, unit=unit,
                source=density_source,
            )

        else:
            # This case should in principle never happen, except if bug in
            # logics of code above
            raise ValueError('Unknown error --  please check code of convert() function.')



"""Main module to calculate the properties of aqueous solutions."""

# TODO: add densities expression of Clegg & Wexler 2011 (eq. 24)
# TODO: add expression of Pitzer 1982 (source of CRC Handbook)
# TODO: write more comprehensive examples


from .convert import convert

from ..properties import SolutionProperty, SolutionSolubilityProperty

from ..formulas.solutions.activity_coefficient import ActivityCoefficientFormulas
from ..formulas.solutions.water_activity import WaterActivityFormulas
from ..formulas.solutions.density import DensityFormulas
from ..formulas.solutions.surface_tension import SurfaceTensionFormulas
from ..formulas.solutions.electrical_conductivity import ElectricalConductivityFormulas
from ..formulas.solutions.refractive_index import RefractiveIndexFormulas
from ..formulas.solutions.solubility import SolubilityFormulas


class SolutionProperty_Full(SolutionProperty):
    """Solution property with full converter (including molarity).

    Is used to prevent circular import problems
    (because SolutionProperty is also used to define the density function used
    in convert())
    """
    # See SolutionProperty for explanation of necessity of staticmethod()
    converter = staticmethod(convert)


class ActivityCoefficient(SolutionProperty_Full):
    """Molal activity coefficient (Ɣ) of solute in a solution (a_s = Ɣ * m / mref)

    Examples
    --------
    - activity_coefficient(m=6.1)  # at saturation for NaCl
    - activity_coefficient(solute='KCl', T=50, m=[2, 4, 6])  # concentration as iterable
    """
    Formulas = ActivityCoefficientFormulas
    quantity = 'activity coefficient'
    unit = '[-]'


class WaterActivity(SolutionProperty_Full):
    """Water activity of a solution(aq) at given concentration and temperature

    Examples
    --------
    - water_activity(x=0.1) returns a_w for a mole fraction of 0.1 of NaCl
    - water_activity(w=0.2) returns a_w for a mass fraction of 0.2 of NaCl
    - water_activity(c=5000) returns a_w for a molality of 5 mol/L of NaCl
    - water_activity(m=6) returns a_w for a molality of 6 mol/kg of NaCl
    - water_activity('LiCl', m=6): same for LiCl
    - water_activity('LiCl', m=6, T=30): same for LiCl at 30°C
    - water_activity('LiCl', 293, 'K', m=6): same for LiCl at 293K.
    - water_activity(solute='CaCl2', T=50, m=[2, 4, 6])  # concentration as iterable
    """
    Formulas = WaterActivityFormulas
    quantity = 'water activity'
    unit = '[-]'


class Density(SolutionProperty_Full):
    """Density of a solution(aq) at a given concentration and temperature

    Examples
    --------
    - density(w=0.1) returns the density of a NaCl solution, calculated with
    Simion equation for a mass fraction of 0.1 at a temperature of 25°C.
    - density('LiCl', 300, 'K', m=6) density of a LiCl solution at 300K
    for a molality of 6 mol/kg.
    - density(source='Tang', x=0.1), density of NaCl solution at a mole
    fraction of 0.1, calculated with the equation from Tang.
    - density(c=5000, relative=True), relative density of NaCl solution at
    a concentration of 5 mol/L.
    """
    Formulas = DensityFormulas
    quantity = 'density'
    unit = '[kg/m^3]'


class SurfaceTension(SolutionProperty_Full):
    """Surface tension of a solution(aq) at a given concentration and temperature

    Examples
    --------
    - surface_tension(x=0.05) returns surface tension of an aqueous NaCl
    solution at 25°C and a mole fraction of 5%
    - surface_tension('LiCl', w=0.1) returns the surface tension of a LiCl
    solution at 25°C and weight fraction of 10%
    - surface_tension('CaCl2', 20, m=6) returns the surface tension of
    a CaCl2 solution at 20°C and molality 6 mol/kg
    - surface_tension('CaCl2', 300, 'K', c=5e3) returns the surface tension of
    a CaCl2 solution at 300K and molarity of 5 mol/L
    - surface_tension(x=[0.02, 0.04, 0.08])  # iterable concentration is ok
    """
    Formulas = SurfaceTensionFormulas
    quantity = 'surface tension'
    unit = '[N/m]'


class ElectricalConductivity(SolutionProperty_Full):
    """Electrical conductivity of an aqueous solution at a given concentration.

    Examples
    --------
    - electrical_conductivity(c=1000)  # 1 molar NaCl conductivity
    - electrical_conductivity(solute='KCl', m=0.1)
    - electrical_conductivity(solute='KCl, m=2.2, T=50)  # at 50°C

    (Note: arrays are accepted for concentration and temperature)
    """
    Formulas = ElectricalConductivityFormulas
    quantity = 'electrical conductivity'
    unit = '[S/m]'


class RefractiveIndex(SolutionProperty_Full):
    """Refractive index of a solution as a function of concentration and temperature

    Examples
    --------
    - refractive_index(x=0.1) returns n for a mole fraction of 0.1 of NaCl
    - refractive_index(w=0.2) returns n for a mass fraction of 0.2 of NaCl
    - refractive_index(c=5000) returns n for a molality of 5 mol/L of NaCl
    - refractive_index(m=3) returns n for a molality of 6 mol/kg of NaCl
    - refractive_index('KCl', m=3): same for KCl
    - refractive_index('KCl', m=3, T=30): same for KCl at 30°C
    - refractive_index('KCl', 293, 'K', m=3): same for KCl at 293K.
    """
    Formulas = RefractiveIndexFormulas
    quantity = 'refractive index'
    unit = '[-]'


class Solubility(SolutionSolubilityProperty):
    """Solubility as a function of temperature.

    Examples
    --------
    - solubility()                   # solubility (molality) of NaCl at 25°C
    - solubility(T=40)               # solubility (molality) of NaCl at 40°C
    - solubility(T=40, out='x')      # same, but in terms of mole fraction
    - solubility(T=40, out='c')      # same, but in terms of molarity (mol/m^3)
    - solubility('KCl', T=303.15, unit='K')  # solubility of KCl at 30°C
    - solubility(T=[0, 10, 20, 30])          # iterables accepted too
    - solubility('Na2SO4')           # solubility of Na2SO4 at 25°C
    """
    Formulas = SolubilityFormulas
    quantity = 'solubility'
    converter = staticmethod(convert)


# ================ GENERATE USABLE OBJECTS FROM ABOVE CLASSES ================

activity_coefficient = ActivityCoefficient()
water_activity = WaterActivity()
density = Density()
surface_tension = SurfaceTension()
electrical_conductivity = ElectricalConductivity()
refractive_index = RefractiveIndex()
solubility = Solubility()


# # ================================== ACTIVITY ================================


# def activity_coefficient(solute='NaCl', T=25, unit='C', source=None, **concentration):
#     """Molal activity coefficient (Ɣ) of solute in a solution (a_s = Ɣ * m / mref)

#     Parameters
#     ----------
#     - solute (str): solute name, default 'NaCl'
#     - T (float): temperature (default 25)
#     - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin

#     - source (str, default None) : Source for the used equation, if None then
#     gets the default source for the particular solute (defined in submodules).
#     See summary of available sources below.

#     - **concentration: kwargs with any unit that is allowed by convert(), e.g.
#         - m= : molality (mol/kg)
#         - w= : mass fraction
#         - x= : mole fraction
#         - c= : molarity (mol/m^3)
#         - r= : mass ratio (unitless)

#     Output
#     ------
#     - Activity coefficient of solute (dimensionless)

#     Solutes and Sources
#     -------------------
#     (* = default)
#     KCl: 'Steiger 2008'(*), 'Tang'
#     Na2SO4: 'Steiger 2005', 'Steiger 2008'(*)
#     NaCl (default solute): 'Steiger 2005', 'Steiger 2008'(*), 'Tang'

#     See details about the sources in the submodules and Readme file.

#     Examples
#     --------
#     - activity_coefficient(m=6.1)  # at saturation for NaCl
#     - activity_coefficient(solute='KCl', T=50, m=[2, 4, 6])  # concentration as iterable
#     """
#     gamma = calculation(
#         propty='activity coefficient',
#         solute=solute,
#         source=source,
#         parameters=(T, unit, concentration),
#         converter=convert,
#     )
#     return format_output_type(gamma)


# def water_activity(solute='NaCl', T=25, unit='C', source=None, **concentration):
#     """Return water activity of an aqueous solution at a given concentration.

#     Parameters
#     ----------
#     - solute (str): solute name, default 'NaCl'
#     - T (float): temperature (default 25)
#     - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin

#     - source (str, default None) : Source for the used equation, if None then
#     gets the default source for the particular solute (defined in submodules).
#     See summary of available sources below.

#     - **concentration: kwargs with any unit that is allowed by convert(), e.g.
#         - m= : molality (mol/kg)
#         - w= : mass fraction
#         - x= : mole fraction
#         - c= : molarity (mol/m^3)
#         - r= : mass ratio (unitless)

#     Output
#     ------
#     - Water activity (range 0-1)

#     Solutes and Sources
#     -------------------
#     (* = default)
#     CaCl2: 'Conde'(*)
#     KCl: 'Steiger 2008'(*), 'Tang'
#     LiCl: 'Conde'(*)
#     Na2SO4: 'Clegg', 'Steiger 2005', 'Steiger 2008'(*)
#     NaCl (default solute): 'Clegg' (*), 'Tang', 'Steiger 2005', 'Steiger 2008'

#     See details about the sources in the submodules and Readme file.

#     Examples
#     --------
#     - water_activity(x=0.1) returns a_w for a mole fraction of 0.1 of NaCl
#     - water_activity(w=0.2) returns a_w for a mass fraction of 0.2 of NaCl
#     - water_activity(c=5000) returns a_w for a molality of 5 mol/L of NaCl
#     - water_activity(m=6) returns a_w for a molality of 6 mol/kg of NaCl
#     - water_activity('LiCl', m=6): same for LiCl
#     - water_activity('LiCl', m=6, T=30): same for LiCl at 30°C
#     - water_activity('LiCl', 293, 'K', m=6): same for LiCl at 293K.
#     - water_activity(solute='CaCl2', T=50, m=[2, 4, 6])  # concentration as iterable
#     """
#     a_w = calculation(
#         propty='water activity',
#         solute=solute,
#         source=source,
#         parameters=(T, unit, concentration),
#         converter=convert,
#     )
#     return format_output_type(a_w)


# # =================================== DENSITY ================================


# def density(solute='NaCl', T=25, unit='C', relative=False, source=None, **concentration):
#     """Return the density of an aqueous solution at a given concentration.

#     Parameters
#     ----------
#     - solute (str): solute name, default 'NaCl'
#     - T (float): temperature (default 25)
#     - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
#     - relative (bool, default False): True for relative density

#     - source (str, default None) : Source for the used equation, if None then
#     gets the default source for the particular solute (defined in submodules).
#     See summary of available sources below.

#     - **concentration: kwargs with any unit that is allowed by convert(), e.g.
#         - m= : molality (mol/kg)
#         - w= : mass fraction
#         - x= : mole fraction
#         - c= : molarity (mol/m^3)
#         - r= : mass ratio (unitless)

#     Output
#     ------
#     - density (kg/m^3) or relative density (dimensionless) if relative is True

#     Solutes and Sources
#     -------------------
#     (* = default)
#     CaCl2: 'Conde'(*), 'Al Ghafri', 'Krumgalz'
#     KCl: 'Al Ghafri'(*), 'Krumgalz'
#     KI: 'Al Ghafri'(*), 'Krumgalz'
#     LiCl: 'Conde'(*), 'Krumgalz'
#     MgCl2: 'Al Ghafri'(*), 'Krumgalz'
#     Na2SO4: 'Tang'(*), 'Clegg', 'Krumgalz'
#     NaCl (default solute): 'Simion'(*), 'Tang', 'Al Ghafri',
#                            'Steiger', 'Krumgalz', 'Clegg'

#     See details about the sources in the submodules and Readme file.

#     Examples
#     --------
#     - density(w=0.1) returns the density of a NaCl solution, calculated with
#     Simion equation for a mass fraction of 0.1 at a temperature of 25°C.
#     - density('LiCl', 300, 'K', m=6) density of a LiCl solution at 300K
#     for a molality of 6 mol/kg.
#     - density(source='Tang', x=0.1), density of NaCl solution at a mole
#     fraction of 0.1, calculated with the equation from Tang.
#     - density(c=5000, relative=True), relative density of NaCl solution at
#     a concentration of 5 mol/L.
#     """
#     rho0, rho = calculation(
#         propty='density',
#         solute=solute,
#         source=source,
#         parameters=(T, unit, concentration),
#         converter=convert,
#     )
#     if relative:
#         return format_output_type(rho / rho0)
#     else:
#         return format_output_type(rho)


# # ============================== SURFACE TENSION =============================


# def surface_tension(solute='NaCl', T=25, unit='C', relative=False, source=None, **concentration):
#     """Surface tension of a solution as a function of concentration and temperature

#     Parameters
#     ----------
#     - solute (str): solute name, default 'NaCl'
#     - T (float): temperature (default 25)
#     - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
#     - relative (bool, default False): True to normalize with pure water at T.

#     - source (str, default None) : Source for the used equation, if None then
#     gets the default source for the particular solute (defined in submodules).
#     See summary of available sources below.

#     - **concentration: kwargs with any unit that is allowed by convert(), e.g.
#         - m= : molality (mol/kg)
#         - w= : mass fraction
#         - x= : mole fraction
#         - c= : molarity (mol/m^3)
#         - r= : mass ratio (unitless)

#     Output
#     ------
#     - sigma (float): surface tension (absolute in N/m or relative).

#     Solutes and Sources
#     -------------------
#     (* = default)
#     CaCl2: 'Dutcher'(*), 'Conde'
#     KCl: 'Dutcher'(*)
#     LiCl: 'Conde'(*)
#     MgCl2: 'Dutcher'(*)
#     Na2SO4: 'Dutcher'(*)
#     NaCl (default solute): 'Dutcher'(*), 'Steiger'

#     See details about the sources in the submodules and Readme file.

#     Examples
#     --------
#     - surface_tension(x=0.05) returns surface tension of an aqueous NaCl
#     solution at 25°C and a mole fraction of 5%
#     - surface_tension('LiCl', w=0.1) returns the surface tension of a LiCl
#     solution at 25°C and weight fraction of 10%
#     - surface_tension('CaCl2', 20, m=6) returns the surface tension of
#     a CaCl2 solution at 20°C and molality 6 mol/kg
#     - surface_tension('CaCl2', 300, 'K', c=5e3) returns the surface tension of
#     a CaCl2 solution at 300K and molarity of 5 mol/L
#     - surface_tension(x=[0.02, 0.04, 0.08])  # iterable concentration is ok
#     """
#     s0, s = calculation(
#         propty='surface tension',
#         solute=solute,
#         source=source,
#         parameters=(T, unit, concentration),
#         converter=convert,
#     )
#     if relative:
#         return format_output_type(s / s0)
#     else:
#         return format_output_type(s)


# # =================================== OPTICS =================================


# def refractive_index(solute='NaCl', T=25, unit='C', source=None, **concentration):
#     """Refractive index of a solution as a function of concentration and temperature

#     Parameters
#     ----------
#     - solute (str): solute name, default 'NaCl'
#     - T (float): temperature (default 25)
#     - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin
#     - source (str, default None) : Source for the used equation, if None then
#     gets the default source for the particular solute (defined in submodules).
#     See summary of available sources below.

#     - **concentration: kwargs with any unit that is allowed by convert(), e.g.
#         - m= : molality (mol/kg)
#         - w= : mass fraction
#         - x= : mole fraction
#         - c= : molarity (mol/m^3)
#         - r= : mass ratio (unitless)

#     Output
#     ------
#     - n (float): refractive index (dimensionless)

#     Solutes and Sources
#     -------------------
#     (* = default)
#     CaCl2: 'Tan'(*),
#     KCl: 'Tan'(*)
#     NaCl (default solute): 'Tan'(*)

#     See details about the sources in the submodules and Readme file.

#     Examples
#     --------
#     - refractive_index(x=0.1) returns n for a mole fraction of 0.1 of NaCl
#     - refractive_index(w=0.2) returns n for a mass fraction of 0.2 of NaCl
#     - refractive_index(c=5000) returns n for a molality of 5 mol/L of NaCl
#     - refractive_index(m=3) returns n for a molality of 6 mol/kg of NaCl
#     - refractive_index('KCl', m=3): same for KCl
#     - refractive_index('KCl', m=3, T=30): same for KCl at 30°C
#     - refractive_index('KCl', 293, 'K', m=3): same for KCl at 293K.
#     """
#     n = calculation(
#         propty='refractive index',
#         solute=solute,
#         source=source,
#         parameters=(T, unit, concentration),
#         converter=convert,
#     )
#     return format_output_type(n)


# # ================================= ELECTRICAL ===============================


# def electrical_conductivity(solute='NaCl', T=25, unit='C', source=None, **concentration):
#     """Return electrical conductivity of an aqueous solution at a given concentration.

#     Parameters
#     ----------
#     - solute (str): solute name, default 'NaCl'
#     - T (float): temperature (default 25)
#     - unit (str, default 'C'): 'C' for Celsius, 'K' for Kelvin

#     - source (str, default None) : Source for the used equation, if None then
#     gets the default source for the particular solute (defined in submodules).
#     See summary of available sources below.

#     - **concentration: kwargs with any unit that is allowed by convert(), e.g.
#         - m= : molality (mol/kg)
#         - w= : mass fraction
#         - x= : mole fraction
#         - c= : molarity (mol/m^3)
#         - r= : mass ratio (unitless)

#     Output
#     ------
#     - Electrical conductivity (S/m)

#     Solutes and Sources
#     -------------------
#     (* = default)
#     KCl: 'McKee'(*)
#     NaCl (default solute): 'Sinmyo'(*)

#     See details about the sources in the submodules and Readme file.

#     Examples
#     --------
#     - electrical_conductivity(c=1000)  # 1 molar NaCl conductivity
#     - electrical_conductivity(solute='KCl', m=0.1)
#     - electrical_conductivity(solute='KCl, m=2.2, T=50)  # at 50°C

#     (Note: arrays are accepted for concentration and temperature)
#     """
#     sigma = calculation(
#         propty='electrical conductivity',
#         solute=solute,
#         source=source,
#         parameters=(T, unit, concentration),
#         converter=convert,
#     )
#     return format_output_type(sigma)



"""Solubility of salts from the CRC Handbook.

Based on polynomial fits of the CRC Handbook data

Sources
-------
- CRC Handbook of Chemistry and Physics, 104th Edition
  Section: 4 | Solubility of Common Inorganic Salts as a Function of Temperature
  https://hbcp.chemnetbase.com/documents/04_29/04_29_0001.xhtml?dswid=7662
"""

import numpy as np

from ....format import make_array_method
from ...general import SaturatedSolutionFormula


class Solubility_CRCHandbook_Base(SaturatedSolutionFormula):

    source = 'CRC Handbook'

    concentration_unit = 'm'
    temperature_unit = 'C'

    def calculate(self, T):
        a0, a1, a2 = self.coeffs
        return a0 + a1 * T + a2 * T**2


# =============================== Steiger 2008 ===============================


class Solubility_NaCl_CRCHandbook_Base(Solubility_CRCHandbook_Base):
    crystal = 'NaCl'
    temperature_range = (10, 40)
    coeffs = 6.09457143, 9.47619048e-4, 5.61904762e-5


class Solubility_LiCl_CRCHandbook_Base(Solubility_CRCHandbook_Base):
    crystal = 'LiCl'
    temperature_range = (10, 25)
    coeffs = 19.13215, 5.87e-03, 1.05e-3
